from sqlalchemy.sql import func

from jobmon.server.web import session_factory
from jobmon.server.web.server_side_exception import ServerError


def add_time(next_report_increment: float) -> func:
    """Adds the next report increment time, and then returns the new time."""
    if session_factory().bind.dialect.name == "mysql":
        add_time_func = func.ADDTIME(
            func.now(), func.SEC_TO_TIME(next_report_increment)
        )
    elif session_factory().bind.dialect.name == "sqlite":
        add_time_func = func.datetime(func.now(), f"+{next_report_increment} seconds")

    else:
        raise ServerError(
            "invalid sql dialect. Only (mysql, sqlite) are supported. Got"
            + session_factory.bind.dialect.name
        )
    return add_time_func


def subtract_time(next_report_increment: float) -> func:
    """Subtracts the next report increment time, and then returns the new time."""
    if session_factory().bind.dialect.name == "mysql":
        add_time_func = func.SUBTIME(
            func.now(), func.SEC_TO_TIME(next_report_increment)
        )
    elif session_factory().bind.dialect.name == "sqlite":
        add_time_func = func.datetime(func.now(), f"-{next_report_increment} seconds")

    else:
        raise ServerError(
            "invalid sql dialect. Only (mysql, sqlite) are supported. Got"
            + session_factory.bind.dialect.name
        )
    return add_time_func



"""Configure Logging for structlogs, syslog, etc."""
import logging.config
from typing import Any, Dict, MutableMapping, Optional

from elasticapm.handlers.structlog import structlog_processor as elasticapm_processor
import structlog

from jobmon.server import __version__


def _processor_add_version(
    logger: logging.Logger, log_method: str, event_dict: MutableMapping[str, Any]
) -> MutableMapping[str, Any]:
    event_dict["jobmon_version"] = __version__
    return event_dict


def configure_structlog() -> None:
    """Configure logging format, handlers, etc."""
    structlog.configure(
        processors=[
            # bring in threadlocal context
            structlog.contextvars.merge_contextvars,
            # This performs the initial filtering, so we don't
            # evaluate e.g. DEBUG when unnecessary
            structlog.stdlib.filter_by_level,
            # Adds logger=module_name (e.g __main__)
            structlog.stdlib.add_logger_name,
            # Adds level=info, debug, etc.
            structlog.stdlib.add_log_level,
            # add jobmon version to json object
            _processor_add_version,
            # Performs the % string interpolation as expected
            structlog.stdlib.PositionalArgumentsFormatter(),
            # add datetime to our logs
            structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S"),
            # Include the stack when stack_info=True
            structlog.processors.StackInfoRenderer(),
            # Adds transaction.id, trace.id, span.id for APM visualization
            elasticapm_processor,
            # Include the exception when exc_info=True
            # e.g log.exception() or log.warning(exc_info=True)'s behavior
            structlog.processors.format_exc_info,
            # Creates the necessary args, kwargs for log()
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        # Our "event_dict" is explicitly a dict
        context_class=dict,
        # Provides the logging.Logger for the underlaying log call
        logger_factory=structlog.stdlib.LoggerFactory(),
        # Provides predefined methods - log.debug(), log.info(), etc.
        wrapper_class=structlog.stdlib.BoundLogger,
        # Caching of our logger
        cache_logger_on_first_use=True,
    )


default_formatters: Dict = {
    # copied formatter from here: https://github.com/hynek/structlog/issues/235
    "text": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.dev.ConsoleRenderer(),
        "keep_exc_info": True,
        "keep_stack_info": True,
    },
    "json": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.processors.JSONRenderer(),
    },
}
default_handlers: Dict = {
    "console_text": {
        "level": "INFO",
        "class": "logging.StreamHandler",
        "formatter": "text",
    },
    "console_json": {
        "level": "INFO",
        "class": "logging.StreamHandler",
        "formatter": "json",
    },
}
default_loggers: Dict = {
    # only configure loggers of given name
    "jobmon.server.web": {
        "handlers": ["console_json"],
        "level": "INFO",
    },
    "werkzeug": {
        "handlers": ["console_json"],
        "level": "WARN",
    },
    "sqlalchemy": {
        "handlers": ["console_json"],
        "level": "WARN",
    }
    # enable SQL debug
    # 'sqlalchemy.engine': {
    #     'level': 'INFO',
    # }
}


def configure_logging(
    loggers_dict: Optional[Dict] = None,
    handlers_dict: Optional[Dict] = None,
    formatters_dict: Optional[Dict] = None,
) -> None:
    """Setup logging with default handlers."""
    if formatters_dict is None:
        formatters_dict = default_formatters
    if handlers_dict is None:
        handlers_dict = default_handlers
    if loggers_dict is None:
        loggers_dict = default_loggers
    dict_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": formatters_dict,
        "handlers": handlers_dict,
        "loggers": loggers_dict,
    }
    logging.config.dictConfig(dict_config)



"""Routes used by the main jobmon client."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict

from flask import jsonify, request
from sqlalchemy import insert, select
import structlog


from jobmon.server.web.models.node import Node
from jobmon.server.web.models.node_arg import NodeArg
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint


logger = structlog.get_logger(__name__)


@blueprint.route("/nodes", methods=["POST"])
def add_nodes() -> Any:
    """Add a chunk of nodes to the database.

    Args:
        nodes: a list of
            node_args_hash: unique identifier of all NodeArgs associated with a node.
        task_template_version_id: version id of the task_template a node belongs to.
        node_args: key-value pairs of arg_id and a value.
    """
    data = cast(Dict, request.get_json())
    # Extract node and node_args

    # Bulk insert the nodes and node args with raw SQL, for performance. Ignore duplicate
    # keys
    session = SessionLocal()
    with session.begin():
        node_keys = [
            (n["task_template_version_id"], n["node_args_hash"]) for n in data["nodes"]
        ]
        node_insert_stmt = insert(Node).values(
            [
                {"task_template_version_id": ttv, "node_args_hash": arghash}
                for ttv, arghash in node_keys
            ]
        )
        if SessionLocal.bind.dialect.name == "mysql":
            node_insert_stmt = node_insert_stmt.prefix_with("IGNORE")
        if SessionLocal.bind.dialect.name == "sqlite":
            node_insert_stmt = node_insert_stmt.prefix_with("OR IGNORE")

        session.execute(node_insert_stmt)
        session.flush()

        # Retrieve the node IDs
        ttvids, node_arg_hashes = zip(*node_keys)
        select_stmt = select(Node).where(
            Node.task_template_version_id.in_(ttvids),
            Node.node_args_hash.in_(node_arg_hashes),
        )
        nodes = session.execute(select_stmt).scalars().all()

        node_id_dict = {
            (n.task_template_version_id, n.node_args_hash): n.id for n in nodes
        }

        # Add node args. Cast hash to string to match DB schema
        node_args = {
            (n["task_template_version_id"], n["node_args_hash"]): n["node_args"]
            for n in data["nodes"]
        }

        node_args_list = []
        for node_id_tuple, arg in node_args.items():
            node_id = node_id_dict[node_id_tuple]
            local_logger = logger.bind(node_id=node_id)

            for arg_id, val in arg.items():
                local_logger.debug(
                    "Adding node_arg", node_id=node_id, arg_id=arg_id, val=val
                )
                node_args_list.append(
                    {"node_id": node_id, "arg_id": arg_id, "val": val}
                )

        # Bulk insert again with raw SQL
        if node_args_list:
            node_arg_insert_stmt = insert(NodeArg).values(node_args_list)
            if SessionLocal.bind.dialect.name == "mysql":
                node_arg_insert_stmt = node_arg_insert_stmt.prefix_with("IGNORE")
            if SessionLocal.bind.dialect.name == "sqlite":
                node_arg_insert_stmt = node_arg_insert_stmt.prefix_with("OR IGNORE")

            session.execute(node_arg_insert_stmt)
            session.flush()

    # return result
    return_nodes = {
        ":".join(str(i) for i in key): val for key, val in node_id_dict.items()
    }
    resp = jsonify(nodes=return_nodes)
    resp.status_code = StatusCodes.OK
    return resp



"""Routes for Task Resources."""
import ast
from http import HTTPStatus as StatusCodes
from typing import Any

from flask import jsonify
from sqlalchemy import select
import structlog

from jobmon.server.web.models.queue import Queue
from jobmon.server.web.models.task_resources import TaskResources
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint


logger = structlog.get_logger(__name__)


@blueprint.route("/task_resources/<task_resources_id>", methods=["POST"])
def get_task_resources(task_resources_id: int) -> Any:
    """Return an task_resources."""
    structlog.contextvars.bind_contextvars(task_resources_id=task_resources_id)

    session = SessionLocal()
    with session.begin():
        select_stmt = (
            select(TaskResources.requested_resources, Queue.name)
            .join_from(TaskResources, Queue, TaskResources.queue_id == Queue.id)
            .where(TaskResources.id == task_resources_id)
        )
        requested_resources, queue_name = session.execute(select_stmt).fetchone()
        requested_resources = ast.literal_eval(requested_resources)

    resp = jsonify(requested_resources=requested_resources, queue_name=queue_name)
    resp.status_code = StatusCodes.OK
    return resp



"""Start up workflow reaper service."""
from typing import Callable, Optional


from jobmon.core.exceptions import ConfigError
from jobmon.core.requester import Requester
from jobmon.server.workflow_reaper.notifiers import SlackNotifier
from jobmon.server.workflow_reaper.workflow_reaper import WorkflowReaper


def start_workflow_reaper(
    service_url: str = "",
    slack_api_url: str = "",
    slack_token: str = "",
    slack_channel_default: str = "",
    poll_interval_minutes: Optional[int] = None,
) -> None:
    """Start monitoring for lost workflow runs."""
    # build slack notifier
    wf_sink: Optional[Callable[[str, Optional[str]], None]] = None
    if slack_api_url or slack_token or slack_channel_default:
        try:
            wf_notifier = SlackNotifier(
                api_url=slack_api_url,
                token=slack_token,
                channel_default=slack_channel_default,
            )
            wf_sink = wf_notifier.send
        except ConfigError:
            pass

    # construct requester
    requester: Optional[Requester] = None
    if service_url:
        requester = Requester(service_url)

    poll_interval_seconds: Optional[int] = None
    if poll_interval_minutes is not None:
        poll_interval_seconds = poll_interval_minutes * 60

    reaper = WorkflowReaper(
        poll_interval_seconds=poll_interval_seconds,
        requester=requester,
        wf_notification_sink=wf_sink,
    )
    reaper.monitor_forever()



"""Routes for Tasks."""
from http import HTTPStatus as StatusCodes
import json
from typing import Any, cast, Dict, List, Set

from flask import jsonify, request
import pandas as pd
from sqlalchemy import func, select, update
from sqlalchemy.orm import Session
import structlog

from jobmon.core import constants
from jobmon.core.constants import Direction
from jobmon.core.serializers import SerializeTaskResourceUsage
from jobmon.server.web.models.edge import Edge
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_instance import TaskInstance
from jobmon.server.web.models.task_instance_error_log import TaskInstanceErrorLog
from jobmon.server.web.models.task_instance_status import TaskInstanceStatus
from jobmon.server.web.models.workflow import Workflow
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.cli import blueprint
from jobmon.server.web.server_side_exception import InvalidUsage

# new structlog logger per flask request context. internally stored as flask.g.logger
logger = structlog.get_logger(__name__)

_task_instance_label_mapping = {
    "Q": "PENDING",
    "B": "PENDING",
    "I": "PENDING",
    "R": "RUNNING",
    "E": "FATAL",
    "Z": "FATAL",
    "W": "FATAL",
    "U": "FATAL",
    "K": "FATAL",
    "D": "DONE",
}

_reversed_task_instance_label_mapping = {
    "PENDING": ["Q", "B", "I"],
    "RUNNING": ["R"],
    "FATAL": ["E", "Z", "W", "U", "K"],
    "DONE": ["D"],
}


@blueprint.route("/task_status", methods=["GET"])
def get_task_status() -> Any:
    """Get the status of a task."""
    task_ids = request.args.getlist("task_ids")
    if len(task_ids) == 0:
        raise InvalidUsage(f"Missing {task_ids} in request", status_code=400)

    # status is an optional arg
    status_request = request.args.getlist("status", None)

    session = SessionLocal()
    with session.begin():
        query_filter = [
            Task.id == TaskInstance.task_id,
            TaskInstanceStatus.id == TaskInstance.status,
        ]
        if status_request:
            if len(status_request) > 0:
                status_codes = [
                    i
                    for arg in status_request
                    for i in _reversed_task_instance_label_mapping[arg]
                ]
            query_filter.append(
                TaskInstance.status.in_(
                    [i for arg in status_request for i in status_codes]
                )
            )

        if task_ids:
            query_filter.append(Task.id.in_(task_ids))
        sql = (
            select(
                Task.id,
                Task.status,
                TaskInstance.id,
                TaskInstance.distributor_id,
                TaskInstanceStatus.label,
                TaskInstance.usage_str,
                TaskInstance.stdout,
                TaskInstance.stderr,
                TaskInstanceErrorLog.description,
            )
            .join_from(
                TaskInstance,
                TaskInstanceErrorLog,
                TaskInstance.id == TaskInstanceErrorLog.task_instance_id,
                isouter=True,
            )
            .where(*query_filter)
        )
        rows = session.execute(sql).all()

    column_names = (
        "TASK_ID",
        "task_status",
        "TASK_INSTANCE_ID",
        "DISTRIBUTOR_ID",
        "STATUS",
        "RESOURCE_USAGE",
        "STDOUT",
        "STDERR",
        "ERROR_TRACE",
    )
    if rows and len(rows) > 0:
        # assign to dataframe for serialization
        df = pd.DataFrame(rows, columns=column_names)
        # remap to jobmon_cli statuses
        df.STATUS.replace(to_replace=_task_instance_label_mapping, inplace=True)
        resp = jsonify(task_instance_status=df.to_json())
    else:
        df = pd.DataFrame({}, columns=column_names)
        resp = jsonify(task_instance_status=df.to_json())

    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task/subdag", methods=["POST"])
def get_task_subdag() -> Any:
    """Used to get the sub dag  of a given task.

    It returns a list of sub tasks as well as a list of sub nodes.
    """
    # Only return sub tasks in the following status. If empty or None, return all
    data = cast(Dict, request.get_json())
    task_ids = data.get("task_ids", [])
    task_status = data.get("task_status", [])

    if not task_ids:
        raise InvalidUsage(f"Missing {task_ids} in request", status_code=400)
    if task_status is None:
        task_status = []
    session = SessionLocal()
    with session.begin():
        select_stmt = (
            select(
                Task.workflow_id.label("workflow_id"),
                Workflow.dag_id.label("dag_id"),
                func.group_concat(Task.node_id).label("node_ids"),
            )
            .join_from(Task, Workflow, Task.workflow_id == Workflow.id)
            .where(Task.id.in_(task_ids))
            .group_by(Task.workflow_id, Workflow.dag_id)
        )
        result = session.execute(select_stmt).one_or_none()

        if not result:
            # return empty values when task_id does not exist or db out of consistency
            resp = jsonify(workflow_id=None, sub_task=None)
            resp.status_code = StatusCodes.OK
            return resp

        # Since we have validated all the tasks belong to the same wf in status_command before
        # this call, assume they all belong to the same wf.
        workflow_id = result.workflow_id
        dag_id = result.dag_id
        node_ids = [int(node_id) for node_id in result.node_ids.split(",")]
        sub_dag_tree = _get_subdag(node_ids, dag_id, session)
        sub_task_tree = _get_tasks_from_nodes(
            workflow_id, sub_dag_tree, task_status, session
        )

    resp = jsonify(workflow_id=workflow_id, sub_task=sub_task_tree)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task/update_statuses", methods=["PUT"])
def update_task_statuses() -> Any:
    """Update the status of the tasks."""
    data = cast(Dict, request.get_json())
    try:
        task_ids = data["task_ids"]
        new_status = data["new_status"]
        workflow_status = data["workflow_status"]
        workflow_id = data["workflow_id"]
    except KeyError as e:
        raise InvalidUsage(
            f"problem with {str(e)} in request to {request.path}", status_code=400
        ) from e

    session = SessionLocal()
    with SessionLocal.begin():
        try:
            update_stmt = update(Task).where(
                Task.id.in_(task_ids), Task.status != new_status
            )
            vals = {"status": new_status}
            session.execute(update_stmt.values(**vals))

            # If job is supposed to be rerun, set task instances to "K"
            if new_status == constants.TaskStatus.REGISTERING:
                task_instance_update_stmt = update(TaskInstance).where(
                    TaskInstance.task_id.in_(task_ids),
                    TaskInstance.status.notin_(
                        [
                            constants.TaskInstanceStatus.ERROR_FATAL,
                            constants.TaskInstanceStatus.DONE,
                            constants.TaskInstanceStatus.ERROR,
                            constants.TaskInstanceStatus.UNKNOWN_ERROR,
                            constants.TaskInstanceStatus.RESOURCE_ERROR,
                            constants.TaskInstanceStatus.KILL_SELF,
                            constants.TaskInstanceStatus.NO_DISTRIBUTOR_ID,
                        ]
                    ),
                )
                vals = {"status": constants.TaskInstanceStatus.KILL_SELF}
                session.execute(task_instance_update_stmt.values(**vals))

                # If workflow is done, need to set it to an error state before resume
                if workflow_status == constants.WorkflowStatus.DONE:
                    workflow_update_stmt = update(Workflow).where(
                        Workflow.id == workflow_id
                    )
                    vals = {"status": constants.WorkflowStatus.FAILED}
                    session.execute(workflow_update_stmt.values(**vals))

            session.commit()
        except KeyError as e:
            session.rollback()
            raise InvalidUsage(
                f"{str(e)} in request to {request.path}", status_code=400
            ) from e

    message = f"updated to status {new_status}"
    resp = jsonify(message)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task_dependencies/<task_id>", methods=["GET"])
def get_task_dependencies(task_id: int) -> Any:
    """Get task's downstream and upstream tasks and their status."""
    session = SessionLocal()
    with session.begin():
        dag_id, workflow_id, node_id = _get_dag_and_wf_id(task_id, session)
        up_nodes = _get_node_dependencies({node_id}, dag_id, session, Direction.UP)
        down_nodes = _get_node_dependencies({node_id}, dag_id, session, Direction.DOWN)
        up_task_dict = _get_tasks_from_nodes(workflow_id, list(up_nodes), [], session)
        down_task_dict = _get_tasks_from_nodes(
            workflow_id, list(down_nodes), [], session
        )
    # return a "standard" json format so that it can be reused by future GUI
    up = (
        []
        if up_task_dict is None or len(up_task_dict) == 0
        else [
            [{"id": k, "status": up_task_dict[k][0], "name": up_task_dict[k][1]}]
            for k in up_task_dict
        ]
    )
    down = (
        []
        if down_task_dict is None or len(down_task_dict) == 0
        else [
            [{"id": k, "status": down_task_dict[k][0], "name": down_task_dict[k][1]}]
            for k in down_task_dict
        ]
    )
    resp = jsonify({"up": up, "down": down})
    resp.status_code = 200
    return resp


@blueprint.route("/tasks_recursive/<direction>", methods=["PUT"])
def get_tasks_recursive(direction: str) -> Any:
    """Get all input task_ids'.

    Either downstream or upsteam tasks based on direction;
    return all recursive(including input set) task_ids in the defined direction.
    """
    direct = constants.Direction.UP if direction == "up" else constants.Direction.DOWN
    data = request.get_json()
    # define task_ids as set in order to eliminate dups
    task_ids = set(data.get("task_ids", []))

    try:
        session = SessionLocal()
        with session.begin():
            tasks_recursive = _get_tasks_recursive(task_ids, direct, session)
        resp = jsonify({"task_ids": list(tasks_recursive)})
        resp.status_code = 200
        return resp
    except InvalidUsage as e:
        raise e


@blueprint.route("/task_resource_usage", methods=["GET"])
def get_task_resource_usage() -> Any:
    """Return the resource usage for a given Task ID."""
    try:
        task_id = int(request.args["task_id"])
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to /task_resource_usage", status_code=400
        ) from e

    session = SessionLocal()
    with SessionLocal.begin():
        select_stmt = (
            select(
                Task.num_attempts,
                TaskInstance.nodename,
                TaskInstance.wallclock,
                TaskInstance.maxpss,
            )
            .join_from(Task, TaskInstance, Task.id == TaskInstance.task_id)
            .where(
                TaskInstance.task_id == task_id,
                TaskInstance.status == constants.TaskInstanceStatus.DONE,
            )
        )
        result = session.execute(select_stmt).one_or_none()

        if result is None:
            resource_usage = SerializeTaskResourceUsage.to_wire(None, None, None, None)
        else:
            resource_usage = SerializeTaskResourceUsage.to_wire(
                result.num_attempts, result.nodename, result.wallclock, result.maxpss
            )

    resp = jsonify(resource_usage)
    resp.status_code = StatusCodes.OK
    return resp


def _get_tasks_recursive(
    task_ids: Set[int], direction: Direction, session: Session
) -> set:
    """Get all input task_ids'.

    Either downstream or upsteam tasks based on direction;
    return all recursive(including input set) task_ids in the defined direction.
    """
    tasks_recursive = set()
    next_nodes = set()
    _workflow_id_first = None
    for task_id in task_ids:
        dag_id, workflow_id, node_id = _get_dag_and_wf_id(task_id, session)
        next_nodes_sub = (
            _get_node_dependencies({node_id}, dag_id, session, Direction.DOWN)
            if direction == constants.Direction.DOWN
            else _get_node_dependencies({node_id}, dag_id, session, Direction.UP)
        )
        if _workflow_id_first is None:
            workflow_id_first = workflow_id
        elif workflow_id != workflow_id_first:
            raise InvalidUsage(
                f"{task_ids} in request belong to different workflow_ids"
                f"({workflow_id_first}, {workflow_id})",
                status_code=400,
            )
        next_nodes.update(next_nodes_sub)

    if len(next_nodes) > 0:
        next_task_dict = _get_tasks_from_nodes(
            workflow_id_first, list(next_nodes), [], session
        )
        if len(next_task_dict) > 0:
            task_recursive_sub = _get_tasks_recursive(
                set(next_task_dict.keys()), direction, session
            )
            tasks_recursive.update(task_recursive_sub)

    tasks_recursive.update(task_ids)

    return tasks_recursive


def _get_dag_and_wf_id(task_id: int, session: Session) -> tuple:
    select_stmt = (
        select(Workflow.dag_id, Task.workflow_id, Task.node_id)
        .join_from(Task, Workflow, Task.workflow_id == Workflow.id)
        .where(Task.id == task_id)
    )
    row = session.execute(select_stmt).one_or_none()

    if row is None:
        return None, None, None
    return int(row.dag_id), int(row["workflow_id"]), int(row["node_id"])


def _get_node_dependencies(
    nodes: set, dag_id: int, session: Session, direction: Direction
) -> Set[int]:
    """Get all upstream nodes of a node.

    Args:
        nodes (set): set of nodes
        dag_id (int): ID of DAG
        session (Session): SQLAlchemy session
        direction (Direction): either up or down
    """
    if direction == Direction.UP:
        select_stmt = select(Edge.upstream_node_ids).where(
            Edge.dag_id == dag_id, Edge.node_id.in_(list(nodes))
        )
    elif direction == Direction.DOWN:
        select_stmt = select(Edge.downstream_node_ids).where(
            Edge.dag_id == dag_id, Edge.node_id.in_(list(nodes))
        )
    else:
        raise ValueError(f"Invalid direction type. Expected one of: {Direction}")
    result = session.execute(select_stmt).all()
    node_ids: Set[int] = set()
    for r in result:
        if r[0] is not None:
            ids = json.loads(r[0])
            node_ids = node_ids.union(set(ids))
    return node_ids


def _get_subdag(node_ids: list, dag_id: int, session: Session) -> list:
    """Get all descendants of a given nodes.

    It only queries the primary keys on the edge table without join.

    Args:
        node_ids (list): list of node IDs
        dag_id (int): ID of DAG
        session (Session): SQLAlchemy sessions
    """
    node_set = set(node_ids)
    node_descendants = node_set
    while len(node_descendants) > 0:
        node_descendants = _get_node_dependencies(
            node_descendants, dag_id, session, Direction.DOWN
        )
        node_set = node_set.union(node_descendants)
    return list(node_set)


def _get_tasks_from_nodes(
    workflow_id: int, nodes: List, task_status: List, session: Session
) -> dict:
    """Get task ids of the given node ids.

    Args:
        workflow_id (int): ID of the workflow
        nodes (list): list of nodes
        task_status (list): list of task statuses
        session (Session): SQLAlchemy session
    """
    if not nodes:
        return {}

    select_stmt = select(Task.id, Task.status, Task.name).where(
        Task.workflow_id == workflow_id, Task.node_id.in_(list(nodes))
    )

    result = session.execute(select_stmt).all()
    task_dict = {}
    for r in result:
        # When task_status not specified, return the full subdag
        if not task_status:
            task_dict[r[0]] = [r[1], r[2]]
        else:
            if r[1] in task_status:
                task_dict[r[0]] = [r[1], r[2]]
    return task_dict


@blueprint.route("/task/get_downstream_tasks", methods=["POST"])
def get_downstream_tasks() -> Any:
    """Get only the direct downstreams of a task."""
    data = cast(Dict, request.get_json())

    task_ids = data["task_ids"]
    dag_id = data["dag_id"]
    session = SessionLocal()
    with session.begin():
        tasks_and_edges = session.execute(
            select(Task.id, Task.node_id, Edge.downstream_node_ids).where(
                Task.id.in_(task_ids),
                Task.node_id == Edge.node_id,
                Edge.dag_id == dag_id,
            )
        ).all()
        result = {
            row.id: [row.node_id, row.downstream_node_ids] for row in tasks_and_edges
        }

    resp = jsonify(downstream_tasks=result)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task/get_ti_details_viz/<task_id>", methods=["GET"])
def get_task_details(task_id: int) -> Any:
    """Get information about TaskInstances associated with specific Task ID."""
    session = SessionLocal()
    with session.begin():
        query = (
            select(
                TaskInstance.id,
                TaskInstanceStatus.label,
                TaskInstance.stdout,
                TaskInstance.stderr,
                TaskInstance.stdout_log,
                TaskInstance.stderr_log,
                TaskInstance.distributor_id,
                TaskInstance.nodename,
                TaskInstanceErrorLog.description,
            )
            .outerjoin_from(
                TaskInstance,
                TaskInstanceErrorLog,
                TaskInstance.id == TaskInstanceErrorLog.task_instance_id,
            )
            .where(
                TaskInstance.task_id == task_id,
                TaskInstance.status == TaskInstanceStatus.id,
            )
        )
        rows = session.execute(query).all()

    column_names = (
        "ti_id",
        "ti_status",
        "ti_stdout",
        "ti_stderr",
        "ti_stdout_log",
        "ti_stderr_log",
        "ti_distributor_id",
        "ti_nodename",
        "ti_error_log_description",
    )
    result = [dict(zip(column_names, row)) for row in rows]
    resp = jsonify(taskinstances=result)
    resp.status_code = 200
    return resp


@blueprint.route("/task/get_task_details_viz/<task_id>", methods=["GET"])
def get_task_details_viz(task_id: int) -> Any:
    """Get status of Task from Task ID."""
    session = SessionLocal()
    with session.begin():
        query = select(
            Task.status,
            Task.workflow_id,
            Task.name,
            Task.command,
            Task.status_date,
        ).where(
            Task.id == task_id,
        )
        rows = session.execute(query).all()

    column_names = (
        "task_status",
        "workflow_id",
        "task_name",
        "task_command",
        "task_status_date",
    )
    result = [dict(zip(column_names, row)) for row in rows]
    resp = jsonify(task_details=result)
    resp.status_code = 200
    return resp



"""ClusterType table in the database."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship, Session

from jobmon.server.web.models import Base


class ClusterType(Base):
    """ClusterType table in the database."""

    __tablename__ = "cluster_type"

    id = Column(Integer, primary_key=True)
    name = Column(String(255), unique=True)
    logfile_templates = Column(String(500))

    # ORM relationships
    clusters = relationship("Cluster", back_populates="cluster_type")


def add_cluster_types(session: Session) -> None:
    """Populate the cluster_type table in the database."""
    cluster_types = [
        ClusterType(name="dummy"),
        ClusterType(name="sequential"),
        ClusterType(name="multiprocess"),
    ]
    session.add_all(cluster_types)



"""Start up the flask services."""
from importlib import import_module
from typing import Any, Dict, List, Optional

from elasticapm.contrib.flask import ElasticAPM
from flask import Flask
import sqlalchemy

from jobmon.core.configuration import JobmonConfig
from jobmon.server.web import session_factory
from jobmon.server.web.hooks_and_handlers import add_hooks_and_handlers


class AppFactory:
    def __init__(
        self,
        config: Optional[JobmonConfig] = None,
        sqlalchemy_database_uri: str = "",
        use_apm: Optional[bool] = None,
        apm_server_url: str = "",
        apm_server_name: str = "",
        apm_server_port: Optional[int] = None,
    ) -> None:
        """Initialization of the App Factory."""
        if config is None:
            config = JobmonConfig()
        # configuration for sqlalchemy
        if not sqlalchemy_database_uri:
            sqlalchemy_database_uri = config.get("db", "sqlalchemy_database_uri")
        self.engine = sqlalchemy.create_engine(
            sqlalchemy_database_uri, pool_recycle=200, future=True
        )

        if use_apm is None:
            use_apm = config.get_boolean("web", "use_apm")
        self.use_apm = use_apm

        if self.use_apm and not apm_server_url:
            apm_server_url = config.get("web", "apm_server_url")
        self.apm_server_url = apm_server_url

        if self.use_apm and not apm_server_name:
            apm_server_name = config.get("web", "apm_server_name")
        self.apm_server_name = apm_server_name

        if self.use_apm and apm_server_port is None:
            apm_server_port = config.get_int("web", "apm_server_port")
        self.apm_server_port = apm_server_port

        # bind the engine to the session factory before importing the blueprint
        session_factory.configure(bind=self.engine)

        # in memory db must be created in same thread as app
        if str(self.engine.url) == "sqlite://":
            from jobmon.server.web.models import init_db

            init_db(self.engine)

    @property
    def flask_config(self) -> Dict[str, Any]:
        flask_config = {}
        if self.use_apm:
            flask_config["ELASTIC_APM"] = {
                # Set the required service name. Allowed characters:
                # a-z, A-Z, 0-9, -, _, and space
                "SERVICE_NAME": self.apm_server_name,
                # Set the custom APM Server URL (default: http://0.0.0.0:8200)
                "SERVER_URL": f"http://{self.apm_server_url}:{self.apm_server_port}",
                # Set the service environment
                "ENVIRONMENT": "development",
                "DEBUG": True,
            }
        return flask_config

    def get_app(
        self, blueprints: List = ["fsm", "cli", "reaper"], url_prefix: str = "/"
    ) -> Flask:
        """Create a Flask app."""
        app = Flask(__name__)
        app.config["CORS_HEADERS"] = "Content-Type"
        app.config.from_mapping(self.flask_config)

        if self.use_apm:
            apm = ElasticAPM(app)
        else:
            apm = None

        # register the blueprints we want. they make use of a scoped session attached
        # to the global session factory
        for blueprint in blueprints:
            mod = import_module(f"jobmon.server.web.routes.{blueprint}")
            app.register_blueprint(getattr(mod, "blueprint"), url_prefix=url_prefix)

        # add request logging hooks
        add_hooks_and_handlers(app, apm)

        return app



"""Routes for Workflow."""
from http import HTTPStatus as StatusCodes
from typing import Any, Dict, Optional

from flask import jsonify, request
from flask_cors import cross_origin
import pandas as pd
from sqlalchemy import func, select, text, update
import structlog

from jobmon.core.constants import WorkflowStatus as Statuses
from jobmon.server.web.models.node import Node
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_template import TaskTemplate
from jobmon.server.web.models.task_template_version import TaskTemplateVersion
from jobmon.server.web.models.tool import Tool
from jobmon.server.web.models.tool_version import ToolVersion
from jobmon.server.web.models.workflow import Workflow
from jobmon.server.web.models.workflow_run import WorkflowRun
from jobmon.server.web.models.workflow_status import WorkflowStatus
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.cli import blueprint


# new structlog logger per flask request context. internally stored as flask.g.logger
logger = structlog.get_logger(__name__)

_cli_label_mapping = {
    "A": "PENDING",
    "G": "PENDING",
    "Q": "PENDING",
    "I": "PENDING",
    "E": "PENDING",
    "O": "SCHEDULED",
    "R": "RUNNING",
    "F": "FATAL",
    "D": "DONE",
}

_reversed_cli_label_mapping = {
    "SCHEDULED": ["O"],
    "PENDING": ["A", "G", "Q", "E", "I"],
    "RUNNING": ["R"],
    "FATAL": ["F"],
    "DONE": ["D"],
}

_cli_order = ["PENDING", "SCHEDULED", "RUNNING", "DONE", "FATAL"]


@blueprint.route("/workflow_validation", methods=["POST"])
@cross_origin()
def get_workflow_validation_status() -> Any:
    """Check if workflow is valid."""
    # initial params
    data = request.get_json()
    task_ids = data["task_ids"]

    # if the given list is empty, return True
    if len(task_ids) == 0:
        resp = jsonify(validation=True)
        resp.status_code = StatusCodes.OK
        return resp

    session = SessionLocal()
    with session.begin():
        # execute query
        query_filter = [Task.workflow_id == Workflow.id, Task.id.in_(task_ids)]
        sql = (
            select(Task.workflow_id, Workflow.status).where(*query_filter)
        ).distinct()
        rows = session.execute(sql).all()
    res = [ti[1] for ti in rows]
    # Validate if all tasks are in the same workflow and the workflow status is dead
    if len(res) == 1 and res[0] in (
        Statuses.FAILED,
        Statuses.DONE,
        Statuses.ABORTED,
        Statuses.HALTED,
    ):
        validation = True
    else:
        validation = False

    resp = jsonify(validation=validation, workflow_status=res[0])
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow/<workflow_id>/workflow_tasks", methods=["GET"])
def get_workflow_tasks(workflow_id: int) -> Any:
    """Get the tasks for a given workflow."""
    limit = request.args.get("limit")
    status_request = request.args.getlist("status", None)
    logger.debug(f"Get tasks for workflow in status {status_request}")

    session = SessionLocal()
    with session.begin():
        if status_request:
            query_filter = [
                Workflow.id == Task.workflow_id,
                Task.status.in_(
                    [
                        i
                        for arg in status_request
                        for i in _reversed_cli_label_mapping[arg]
                    ]
                ),
                Workflow.id == int(workflow_id),
            ]
        else:
            query_filter = [
                Workflow.id == Task.workflow_id,
                Workflow.id == int(workflow_id),
            ]
        sql = (
            select(Task.id, Task.name, Task.status, Task.num_attempts).where(
                *query_filter
            )
        ).order_by(Task.id.desc())
        rows = session.execute(sql).all()
    column_names = ("TASK_ID", "TASK_NAME", "STATUS", "RETRIES")
    res = [dict(zip(column_names, ti)) for ti in rows]
    for r in res:
        r["RETRIES"] = 0 if r["RETRIES"] <= 1 else r["RETRIES"] - 1

    if limit:
        res = res[: int(limit)]

    logger.debug(
        f"The following tasks of workflow are in status {status_request}:\n{res}"
    )
    if res:
        # assign to dataframe for serialization
        df = pd.DataFrame(res, columns=res[0].keys())

        # remap to jobmon_cli statuses
        df.STATUS.replace(to_replace=_cli_label_mapping, inplace=True)
        df = df.to_json()
        resp = jsonify(workflow_tasks=df)
    else:
        df = pd.DataFrame({}, columns=["TASK_ID", "TASK_NAME", "STATUS", "RETRIES"])
        resp = jsonify(workflow_tasks=df.to_json())

    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/workflow/<workflow_id>/validate_username/<username>", methods=["GET"]
)
def get_workflow_user_validation(workflow_id: int, username: str) -> Any:
    """Return all usernames associated with a given workflow_id's workflow runs.

    Used to validate permissions for a self-service request.
    """
    logger.debug(f"Validate user name {username} for workflow")
    session = SessionLocal()
    with session.begin():
        query_filter = [WorkflowRun.workflow_id == workflow_id]
        sql = (select(WorkflowRun.user).where(*query_filter)).distinct()
        rows = session.execute(sql).all()
    usernames = [row[0] for row in rows]

    resp = jsonify(validation=username in usernames)

    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/workflow/<workflow_id>/validate_for_workflow_reset/<username>", methods=["GET"]
)
def get_workflow_run_for_workflow_reset(workflow_id: int, username: str) -> Any:
    """Last workflow_run_id associated with a given workflow_id started by the username.

    Used to validate for workflow_reset:
        1. The last workflow_run of the current workflow must be in error state.
        2. This last workflow_run must have been started by the input username.
        3. This last workflow_run is in status 'E'
    """
    session = SessionLocal()
    with session.begin():
        query_filter = [
            WorkflowRun.workflow_id == workflow_id,
            WorkflowRun.status == "E",
        ]
        sql = (select(WorkflowRun.id, WorkflowRun.user).where(*query_filter)).order_by(
            WorkflowRun.created_date.desc()
        )
        rows = session.execute(sql).all()
    result = None if len(rows) <= 0 else rows[0]
    if result is not None and result[1] == username:
        resp = jsonify({"workflow_run_id": result[0]})
    else:
        resp = jsonify({"workflow_run_id": None})

    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("workflow/<workflow_id>/reset", methods=["PUT"])
def reset_workflow(workflow_id: int) -> Any:
    """Update the workflow's status, all its tasks' statuses to 'G'."""
    session = SessionLocal()
    data = request.get_json()
    partial_reset = data.get("partial_reset", False)
    with session.begin():

        current_time = session.query(func.now()).scalar()

        workflow_query = select(Workflow).where(Workflow.id == workflow_id)
        workflow = session.execute(workflow_query).scalars().one_or_none()
        workflow.reset(current_time=current_time)
        session.flush()

        # Update task statuses associated with the workflow
        # Default behavior is a full workflow reset, all tasks to registered state
        # User can optionally request only a partial reset if they want to resume this workflow
        invalid_statuses = ["G"]
        if partial_reset:
            invalid_statuses.append("D")
        update_filter = [Task.workflow_id == workflow_id, Task.status.notin_(invalid_statuses)]
        update_stmt = (
            update(Task)
            .where(*update_filter)
            .values(status="G", status_date=func.now(), num_attempts=0)
        )
        session.execute(update_stmt)
        session.commit()

    resp = jsonify({})
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow_status", methods=["GET"])
@cross_origin()
def get_workflow_status() -> Any:
    """Get the status of the workflow."""
    # initial params
    params = {}
    user_request = request.args.getlist("user")
    if user_request == "all":  # specifying all is equivalent to None
        user_request = []
    workflow_request = request.args.getlist("workflow_id")
    logger.debug(f"Query for wf {workflow_request} status.")
    if workflow_request == "all":  # specifying all is equivalent to None
        workflow_request = []
    limit = request.args.get("limit")
    # set default to 5 to match status_commands
    limit = int(limit) if limit else 5
    # convert workflow request into sql filter
    if workflow_request:
        workflow_request = [int(w) for w in workflow_request]
        params["workflow_id"] = workflow_request
    else:  # if we don't specify workflow then we use the users
        # convert user request into sql filter
        # directly producing workflow_ids, and thus where_clause
        if user_request:
            session = SessionLocal()
            with session.begin():
                query_filter = [WorkflowRun.user.in_(user_request)]
                sql = (
                    (select(WorkflowRun.workflow_id).where(*query_filter))
                    .distinct()
                    .order_by(WorkflowRun.workflow_id.desc())
                    .limit(limit)
                )
                rows = session.execute(sql).all()
            workflow_request = [int(row[0]) for row in rows]
    # performance improvement one: only query the limited number of workflows
    workflow_request = workflow_request[:limit]
    # performance improvement two: split query
    session = SessionLocal()
    with session.begin():
        query_filter = [
            Workflow.id.in_(workflow_request),
            WorkflowStatus.id == Workflow.status,
        ]
        sql = (
            select(
                Workflow.id, Workflow.name, WorkflowStatus.label, Workflow.created_date
            )
        ).where(*query_filter)
        rows1 = session.execute(sql).all()
    row_map = dict()
    for r in rows1:
        row_map[r[0]] = r
    session = SessionLocal()
    with session.begin():
        query_filter = [
            Task.workflow_id.in_(workflow_request),
        ]
        sql = (
            select(
                Task.workflow_id,
                func.count(Task.status),
                Task.status,
            ).where(*query_filter)
        ).group_by(Task.workflow_id, Task.status)
        rows2 = session.execute(sql).all()

    res = []
    for r in rows2:
        d = dict()
        d["WF_ID"] = r[0]
        d["WF_NAME"] = row_map[r[0]][1]
        d["WF_STATUS"] = row_map[r[0]][2]
        d["TASKS"] = r[1]
        d["STATUS"] = r[2]
        d["CREATED_DATE"] = row_map[r[0]][3]
        session = SessionLocal()
        with session.begin():
            q_filter = [Task.workflow_id == d["WF_ID"], Task.status == d["STATUS"]]
            q = select(Task.num_attempts).where(*q_filter)
            query_result = session.execute(q).all()
        retries = 0
        for rr in query_result:
            retries += 0 if int(rr[0]) <= 1 else int(rr[0]) - 1
        d["RETRIES"] = retries
        res.append(d)
    if res is not None and len(res) > 0:
        # assign to dataframe for aggregation
        df = pd.DataFrame(res, columns=res[0].keys())

        # remap to jobmon_cli statuses
        df.STATUS.replace(to_replace=_cli_label_mapping, inplace=True)

        # aggregate totals by workflow and status
        df = df.groupby(
            ["WF_ID", "WF_NAME", "WF_STATUS", "STATUS", "CREATED_DATE"]
        ).agg({"TASKS": "sum", "RETRIES": "sum"})

        # pivot wide by task status
        tasks = df.pivot_table(
            values="TASKS",
            index=["WF_ID", "WF_NAME", "WF_STATUS", "CREATED_DATE"],
            columns="STATUS",
            fill_value=0,
        )
        for col in _cli_order:
            if col not in tasks.columns:
                tasks[col] = 0
        tasks = tasks[_cli_order]

        # aggregate again without status to get the totals by workflow
        retries = df.groupby(["WF_ID", "WF_NAME", "WF_STATUS", "CREATED_DATE"]).agg(
            {"TASKS": "sum", "RETRIES": "sum"}
        )

        # combine datasets
        df = pd.concat([tasks, retries], axis=1)

        # compute pcts and format
        for col in _cli_order:
            df[col + "_pct"] = (df[col].astype(float) / df["TASKS"].astype(float)) * 100
            df[col + "_pct"] = df[[col + "_pct"]].round(1)
            df[col] = (
                df[col].astype(int).astype(str)
                + " ("
                + df[col + "_pct"].astype(str)
                + "%)"
            )

        # df.replace(to_replace={"0 (0.0%)": "NA"}, inplace=True)
        # final order
        df = df[["TASKS"] + _cli_order + ["RETRIES"]]
        df = df.reset_index()
        df = df.to_json()
        resp = jsonify(workflows=df)
    else:
        df = pd.DataFrame(
            {},
            columns=[
                "WF_ID",
                "WF_NAME",
                "WF_STATUS",
                "CREATED_DATE",
                "TASKS",
                "PENDING",
                "RUNNING",
                "DONE",
                "FATAL",
                "RETRIES",
            ],
        ).to_json()
        resp = jsonify(workflows=df)

    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow_status_viz", methods=["GET"])
def get_workflow_status_viz() -> Any:
    """Get the status of the workflows for GUI."""
    session = SessionLocal()
    wf_ids = request.args.getlist("workflow_ids[]")
    # return DS
    return_dic: Dict[int, Any] = dict()
    for wf_id in wf_ids:
        with session.begin():
            sql = select(
                func.min(Task.num_attempts).label("min"),
                func.max(Task.num_attempts).label("max"),
                func.avg(Task.num_attempts).label("mean"),
            ).where(Task.workflow_id == wf_id)
            attempts = session.execute(sql).all()

        return_dic[int(wf_id)] = {
            "id": int(wf_id),
            "tasks": 0,
            "PENDING": 0,
            "SCHEDULED": 0,
            "RUNNING": 0,
            "DONE": 0,
            "FATAL": 0,
            "MAXC": 0,
            "num_attempts_avg": float(attempts[0]["mean"]),
            "num_attempts_min": int(attempts[0]["min"]),
            "num_attempts_max": int(attempts[0]["max"]),
        }

    with session.begin():
        query_filter = [Task.workflow_id.in_(wf_ids), Task.workflow_id == Workflow.id]
        sql = select(
            Task.workflow_id, Task.status, Workflow.max_concurrently_running
        ).where(*query_filter)
        rows = session.execute(sql).all()

    for row in rows:
        return_dic[row[0]]["tasks"] += 1
        return_dic[row[0]][_cli_label_mapping[row[1]]] += 1
        return_dic[row[0]]["MAXC"] = row[2]

    resp = jsonify(return_dic)
    resp.status_code = 200
    return resp


@blueprint.route("/workflow_overview_viz", methods=["GET"])
def workflows_by_user_form() -> Any:
    """Fetch associated workflows and workflow runs by username."""
    arguments = request.args
    clean_args: Dict[str, Optional[str]] = {}
    for key, value in arguments.items():
        if value is None or value == "undefined" or value == "":
            clean_args[key] = None
        else:
            clean_args[key] = value
    user = arguments.get("user")
    tool = arguments.get("tool")
    wf_name = arguments.get("wf_name")
    wf_args = arguments.get("wf_args")
    wf_attribute = arguments.get("wf_attribute")
    date_submitted = arguments.get("date_submitted")
    status = arguments.get("status")

    session = SessionLocal()
    with session.begin():
        where_clauses = []
        substitution_dict = {}
        if user:
            where_clauses.append("workflow_run.user = :user")
            substitution_dict["user"] = user
        if tool:
            where_clauses.append("tool.name = :tool")
            substitution_dict["tool"] = tool
        if wf_name:
            where_clauses.append("workflow.name = :wf_name")
            substitution_dict["wf_name"] = wf_name
        if wf_args:
            where_clauses.append("workflow.workflow_args = :wf_args")
            substitution_dict["wf_args"] = wf_args
        if wf_attribute:
            where_clauses.append("workflow_attribute.value = :wf_attribute")
            substitution_dict["wf_attribute"] = wf_attribute
        if date_submitted:
            where_clauses.append("workflow.created_date >= :date_submitted")
            substitution_dict["date_submitted"] = date_submitted
        if status:
            where_clauses.append("workflow.status = :status")
            substitution_dict["status"] = status
        inner_where_clause = " AND ".join(where_clauses)
        query = text(
            f"""
            SELECT
                workflow.id,
                workflow.name,
                workflow.created_date,
                workflow.status_date,
                workflow.workflow_args,
                count(distinct workflow_run.id) as num_attempts,
                workflow_status.label,
                tool.name
            FROM
                workflow
                JOIN (
                    SELECT
                        distinct queue_id,
                        workflow_id
                    FROM
                        task
                        JOIN task_resources on task_resources.id = task.task_resources_id
                    WHERE
                        task.workflow_id IN (
                            SELECT
                                workflow_run.workflow_id
                            FROM
                                workflow
                                JOIN tool_version on workflow.tool_version_id = tool_version.id
                                JOIN tool on tool.id = tool_version.tool_id
                                JOIN workflow_run on workflow.id = workflow_run.workflow_id
                                LEFT JOIN workflow_attribute on workflow.id = workflow_attribute.workflow_id
                            WHERE
                                {inner_where_clause}
                        )
                    GROUP BY
                        workflow_id
                ) workflow_queue on workflow.id = workflow_queue.workflow_id
                JOIN queue on queue.id = workflow_queue.queue_id
                JOIN workflow_run on workflow.id = workflow_run.workflow_id
                JOIN tool_version on workflow.tool_version_id = tool_version.id
                JOIN tool on tool.id = tool_version.tool_id
                JOIN workflow_status on workflow.status = workflow_status.id
            WHERE
                cluster_id != 1
            GROUP BY
                workflow.id
            ORDER BY
                workflow.id DESC
    """
        )
        rows = session.execute(query, substitution_dict).all()

    column_names = (
        "wf_id",
        "wf_name",
        "wf_submitted_date",
        "wf_status_date",
        "wf_args",
        "wfr_count",
        "wf_status",
        "wf_tool",
    )
    # Initialize all possible states as 0. No need to return data since it will be refreshed
    # on demand anyways.
    initial_status_counts = {
        label_mapping: 0 for label_mapping in set(_cli_label_mapping.values())
    }
    result = [dict(zip(column_names, row), **initial_status_counts) for row in rows]

    res = jsonify(workflows=result)
    res.return_code = StatusCodes.OK
    return res


@blueprint.route("/task_table_viz/<workflow_id>", methods=["GET"])
def task_details_by_wf_id(workflow_id: int) -> Any:
    """Fetch Task details associated with Workflow ID and TaskTemplate name."""
    task_template_name = request.args.get("tt_name")
    session = SessionLocal()
    with session.begin():
        sql = (
            select(
                Task.id,
                Task.name,
                Task.status,
                Task.command,
                Task.num_attempts,
                Task.status_date,
                Task.max_attempts,
            )
            .where(
                Task.workflow_id == workflow_id,
                Task.node_id == Node.id,
                Node.task_template_version_id == TaskTemplateVersion.id,
                TaskTemplateVersion.task_template_id == TaskTemplate.id,
                TaskTemplate.name == task_template_name,
            )
            .order_by(Task.id.asc())
        )
        rows = session.execute(sql).all()

    column_names = (
        "task_id",
        "task_name",
        "task_status",
        "task_command",
        "task_num_attempts",
        "task_status_date",
        "task_max_attempts",
    )

    result = [dict(zip(column_names, row)) for row in rows]
    for r in result:
        r["task_status"] = _cli_label_mapping[r["task_status"]]

    res = jsonify(tasks=result)
    res.return_code = StatusCodes.OK
    return res


@blueprint.route("/workflow_details_viz/<workflow_id>", methods=["GET"])
def wf_details_by_wf_id(workflow_id: int) -> Any:
    """Fetch name, args, dates, tool for a Workflow provided WF ID."""
    session = SessionLocal()
    with session.begin():
        sql = select(
            Workflow.name,
            Workflow.workflow_args,
            Workflow.created_date,
            Workflow.status_date,
            Tool.name,
            Workflow.status,
            WorkflowStatus.description,
        ).where(
            Workflow.id == workflow_id,
            Workflow.tool_version_id == ToolVersion.id,
            ToolVersion.tool_id == Tool.id,
            WorkflowStatus.id == Workflow.status,
        )
        rows = session.execute(sql).all()

    column_names = (
        "wf_name",
        "wf_args",
        "wf_created_date",
        "wf_status_date",
        "tool_name",
        "wf_status",
        "wf_status_desc",
    )

    result = [dict(zip(column_names, row)) for row in rows]
    resp = jsonify(result)
    resp.status_code = 200
    return resp



"""Routes used to move through the finite state."""
from http import HTTPStatus as StatusCodes
from typing import Any

from flask import jsonify, request
from sqlalchemy import func, select, update
import structlog

from jobmon.core.exceptions import InvalidStateTransition
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.workflow import Workflow
from jobmon.server.web.models.workflow_run import WorkflowRun
from jobmon.server.web.models.workflow_run_status import WorkflowRunStatus
from jobmon.server.web.models.workflow_status import WorkflowStatus
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.reaper import blueprint

# new structlog logger per flask request context. internally stored as flask.g.logger
logger = structlog.get_logger(__name__)


@blueprint.route("/workflow/<workflow_id>/fix_status_inconsistency", methods=["PUT"])
def fix_wf_inconsistency(workflow_id: int) -> Any:
    """Find wf in F with all tasks in D and fix them.

    For flexibility, pass in the step size. It is easier to redeploy the reaper than the
    service.
    """
    data = request.get_json()
    increase_step = data["increase_step"]
    logger.debug(
        f"Fix inconsistencies starting at workflow {workflow_id} by {increase_step}"
    )
    session = SessionLocal()
    with session.begin():
        sql = select(Workflow.id)
        rows = session.execute(sql).all()
        # the id to return to reaper as next start point
        total_wf = len(rows)

    # move the starting row forward by increase_step
    # It takes about 1 second per thousand; increase_step is passed in from the reaper.
    # Lf the starting row > max row, restart from workflow-id 0.
    # This way, we can get to the unfinished the wf later
    # without querying the whole db every time.

    current_max_wf_id = int(workflow_id) + int(increase_step)
    if current_max_wf_id > total_wf:
        logger.info("Fix inconsistencies starting from workflow_id zero again")
        current_max_wf_id = 0

    # Update wf in F with all task in D to D
    # count(s) will have the total number of tasks, sum(s) is those in D.
    # If the two are equal, then the workflow Tasks are all D and therefore the workflow
    # should be D.
    session = SessionLocal()
    with session.begin():
        query_filter = [
            Workflow.id > workflow_id,
            Workflow.id <= int(workflow_id) + increase_step,
            Workflow.status == "F",
            Workflow.id == Task.workflow_id,
        ]
        sql = select(Workflow.id, Task.status).where(*query_filter)
        rows = session.execute(sql).all()
        result_set = set([r[0] for r in rows])
        for r in rows:
            if r[1] != "D" and r[0] in result_set:
                result_set -= {r[0]}
        result_list = list(result_set)

    if result_list is None or len(result_list) == 0:
        logger.debug("No inconsistent F-D workflows to fix.")
    else:
        logger.info("Fixing inconsistent F-D workflow: {ids}")
        session = SessionLocal()
        with session.begin():
            update_stmt = (
                update(Workflow)
                .where(Workflow.id.in_(result_list))
                .values(status="D", status_date=func.now())
            )
            session.execute(update_stmt)
            session.commit()

        logger.debug("Done fixing F-D inconsistent workflows.")

    resp = jsonify({"wfid": current_max_wf_id})
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow/<workflow_id>/workflow_name_and_args", methods=["GET"])
def get_wf_name_and_args(workflow_id: int) -> Any:
    """Return workflow name and args associated with specified workflow ID."""
    session = SessionLocal()
    with session.begin():
        query_filter = [Workflow.id == workflow_id]
        sql = select(Workflow.name, Workflow.workflow_args).where(*query_filter)
        result = session.execute(sql).all()

    if result is None or len(result) == 0:
        # return empty values in case of DB inconsistency
        resp = jsonify(workflow_name=None, workflow_args=None)
        resp.status_code = StatusCodes.OK
        return resp

    resp = jsonify(workflow_name=result[0][0], workflow_args=result[0][1])
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/lost_workflow_run", methods=["GET"])
def get_lost_workflow_runs() -> Any:
    """Return all workflow runs that are currently in the specified state."""
    statuses = request.args.getlist("status")
    version = request.args.get("version")
    session = SessionLocal()
    with session.begin():
        query_filter = [
            WorkflowRun.status.in_(statuses),
            WorkflowRun.heartbeat_date <= func.now(),
            WorkflowRun.jobmon_server_version == version,
        ]
        sql = select(WorkflowRun.id, WorkflowRun.workflow_id).where(*query_filter)
        rows = session.execute(sql).all()
    workflow_runs = [(r[0], r[1]) for r in rows]
    resp = jsonify(workflow_runs=workflow_runs)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow_run/<workflow_run_id>/reap", methods=["PUT"])
def reap_workflow_run(workflow_run_id: int) -> Any:
    """If the last task was more than 2 minutes ago, transition wfr to A state.

    Also check WorkflowRun status_date to avoid possible race condition where reaper
    checks tasks from a different WorkflowRun with the same workflow id. Avoid setting
    while waiting for a resume (when workflow is in suspended state).
    """
    structlog.contextvars.bind_contextvars(workflow_run_id=workflow_run_id)
    logger.info(f"Reap wfr: {workflow_run_id}")

    session = SessionLocal()
    with session.begin():
        # get the wfr
        query_filter = [
            WorkflowRun.id == workflow_run_id,
            WorkflowRun.heartbeat_date <= func.now(),
        ]
        sql = select(WorkflowRun.id, WorkflowRun.workflow_id, WorkflowRun.status).where(
            *query_filter
        )
        rows = session.execute(sql).all()
    if len(rows) == 0:
        resp = jsonify(status="")
        resp.status_code = StatusCodes.OK
        return resp

    # reap wfr
    wfr_id, wf_id, wfr_status = rows[0][0], rows[0][1], rows[0][2]
    if wfr_status == WorkflowRunStatus.LINKING:
        logger.debug(f"Transitioning wfr {wfr_id} to ABORTED")
        target_wfr_status = WorkflowRunStatus.ABORTED
        target_wf_status = WorkflowStatus.ABORTED
    if wfr_status in [WorkflowRunStatus.COLD_RESUME, WorkflowRunStatus.HOT_RESUME]:
        logger.debug(f"Transitioning wfr {wfr_id} to TERMINATED")
        target_wfr_status = WorkflowRunStatus.TERMINATED
        target_wf_status = WorkflowStatus.HALTED
    if wfr_status == WorkflowRunStatus.RUNNING:
        logger.debug(f"Transitioning wfr {wfr_id} to ERROR")
        target_wfr_status = WorkflowRunStatus.ERROR
        target_wf_status = WorkflowStatus.FAILED

    # validate transition
    if (wfr_status, target_wfr_status) not in WorkflowRun().valid_transitions:
        try:
            raise InvalidStateTransition(
                model="WorkflowRun",
                id=wfr_id,
                old_state=wfr_status,
                new_state=target_wfr_status,
            )
        except (InvalidStateTransition, AttributeError) as e:
            # this branch handles race condition or case where no wfr was returned
            logger.debug(f"Unable to reap workflow_run {wfr_id}: {e}")

    # update status
    session = SessionLocal()
    with session.begin():
        query1 = f"""UPDATE workflow_run
                    SET status="{target_wfr_status}"
                    WHERE id={wfr_id}
                """
        session.execute(query1)
        query2 = f"""UPDATE workflow
                     SET status="{target_wf_status}"
                     WHERE id={wf_id}
                """
        session.execute(query2)
        session.commit()

    resp = jsonify(status=target_wfr_status)
    resp.status_code = StatusCodes.OK
    return resp



"""Routes for Tools."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict

from flask import jsonify, request
import sqlalchemy
from sqlalchemy import select
import structlog

from jobmon.server.web.models.tool import Tool
from jobmon.server.web.models.tool_version import ToolVersion
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint
from jobmon.server.web.server_side_exception import InvalidUsage


logger = structlog.get_logger(__name__)


@blueprint.route("/tool", methods=["POST"])
def add_tool() -> Any:
    """Add a tool to the database."""
    data = cast(Dict, request.get_json())
    try:
        tool_name = data["name"]
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    # add tool to db
    session = SessionLocal()
    print(session.bind)
    try:
        with session.begin():
            logger.info(f"Adding tool {tool_name}")
            tool = Tool(name=tool_name)
            session.add(tool)
    except sqlalchemy.exc.IntegrityError:
        with session.begin():
            select_stmt = select(Tool).where(Tool.name == tool_name)
            tool = session.execute(select_stmt).scalars().one()
    wire_format = tool.to_wire_as_client_tool()

    resp = jsonify(tool=wire_format)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/tool/<tool_id>/tool_versions", methods=["GET"])
def get_tool_versions(tool_id: int) -> Any:
    """Get the Tool Version."""
    # check input variable
    structlog.contextvars.bind_contextvars(tool_id=tool_id)
    logger.info(f"Getting available tool versions for tool_id {tool_id}")
    try:
        tool_id = int(tool_id)
    except Exception as e:
        raise InvalidUsage(
            f"Variable tool_id must be an int in {request.path}", status_code=400
        ) from e

    # get data from db
    session = SessionLocal()
    with session.begin():
        select_stmt = select(ToolVersion).where(ToolVersion.tool_id == tool_id)
        tool_versions = session.execute(select_stmt).scalars().all()
    wire_format = [t.to_wire_as_client_tool_version() for t in tool_versions]

    logger.info(f"Tool version for {tool_id} is {wire_format}")
    resp = jsonify(tool_versions=wire_format)
    resp.status_code = StatusCodes.OK
    return resp



"""Task Template database table."""
from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint, VARCHAR
from sqlalchemy.orm import relationship

from jobmon.core.serializers import SerializeClientTaskTemplate
from jobmon.server.web.models import Base


class TaskTemplate(Base):
    """Task Template database table."""

    __tablename__ = "task_template"

    def to_wire_as_client_task_template(self) -> tuple:
        """Serialize Task Template."""
        # serialized = SerializeClientTool.to_wire(id=self.id, name=self.name)
        # return serialized
        return SerializeClientTaskTemplate.to_wire(
            self.id, self.tool_version_id, self.name
        )

    id = Column(Integer, primary_key=True)
    tool_version_id = Column(Integer, ForeignKey("tool_version.id"))
    name = Column(VARCHAR(255))

    # orm relationship
    tool_versions = relationship("ToolVersion", back_populates="task_templates")
    task_template_versions = relationship(
        "TaskTemplateVersion", back_populates="task_template"
    )

    __table_args__ = (
        UniqueConstraint("tool_version_id", "name", name="uc_tool_version_id_name"),
    )



"""Workflow Attribute type database table."""
from typing import Any, Dict

from sqlalchemy import Column, Integer, String

from jobmon.server.web.models import Base


class WorkflowAttributeType(Base):
    """Workflow Attribute type database table."""

    __tablename__ = "workflow_attribute_type"

    id = Column(Integer, primary_key=True)
    name = Column(String(255), unique=True)

    @classmethod
    def from_wire(cls: Any, dct: Dict) -> Any:
        """Workflow attribute object from dict of params."""
        return cls(id=dct["workflow_attribute_type_id"], name=dct["name"])

    def to_wire(self) -> Dict:
        """Workflow attribute attributes to dict."""
        return {"workflow_attribute_type_id": self.id, "name": self.name}



"""Routes for Workflows."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict, List, Tuple

from flask import jsonify, request
import sqlalchemy
from sqlalchemy import func, select, update
from sqlalchemy.dialects.mysql import insert as mysql_insert
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.orm import Session
import structlog

from jobmon.server.web.models.array import Array
from jobmon.server.web.models.dag import Dag
from jobmon.server.web.models.queue import Queue
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_resources import TaskResources
from jobmon.server.web.models.task_status import TaskStatus
from jobmon.server.web.models.workflow import Workflow
from jobmon.server.web.models.workflow_attribute import WorkflowAttribute
from jobmon.server.web.models.workflow_attribute_type import WorkflowAttributeType
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint
from jobmon.server.web.server_side_exception import InvalidUsage


logger = structlog.get_logger(__name__)


def _add_workflow_attributes(
    workflow_id: int, workflow_attributes: Dict[str, str], session: Session
) -> None:
    # add attribute
    structlog.contextvars.bind_contextvars(workflow_id=workflow_id)
    logger.info(f"Add Attributes: {workflow_attributes}")
    wf_attributes_list = []
    with session.begin_nested():
        for name, val in workflow_attributes.items():
            wf_type_id = _add_or_get_wf_attribute_type(name, session)
            wf_attribute = WorkflowAttribute(
                workflow_id=workflow_id,
                workflow_attribute_type_id=wf_type_id,
                value=val,
            )
            wf_attributes_list.append(wf_attribute)
            logger.debug(f"Attribute name: {name}, value: {val}")
        session.add_all(wf_attributes_list)


@blueprint.route("/workflow", methods=["POST"])
def bind_workflow() -> Any:
    """Bind a workflow to the database."""
    try:
        data = cast(Dict, request.get_json())
        tv_id = int(data["tool_version_id"])
        dag_id = int(data["dag_id"])
        whash = str(data["workflow_args_hash"])
        thash = str(data["task_hash"])
        description = data["description"]
        name = data["name"]
        workflow_args = data["workflow_args"]
        max_concurrently_running = data["max_concurrently_running"]
        workflow_attributes = data["workflow_attributes"]

    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    structlog.contextvars.bind_contextvars(
        dag_id=dag_id,
        tool_version_id=tv_id,
        workflow_args_hash=str(whash),
        task_hash=str(thash),
    )
    logger.info("Bind workflow")
    session = SessionLocal()
    with session.begin():
        select_stmt = select(Workflow).where(
            Workflow.tool_version_id == tv_id,
            Workflow.dag_id == dag_id,
            Workflow.workflow_args_hash == whash,
            Workflow.task_hash == thash,
        )
        workflow = session.execute(select_stmt).scalars().one_or_none()
        if workflow is None:
            # create a new workflow
            workflow = Workflow(
                tool_version_id=tv_id,
                dag_id=dag_id,
                workflow_args_hash=whash,
                task_hash=thash,
                description=description,
                name=name,
                workflow_args=workflow_args,
                max_concurrently_running=max_concurrently_running,
            )
            session.add(workflow)
            session.flush()
            logger.info("Created new workflow")

            # update attributes
            if workflow_attributes:
                _add_workflow_attributes(workflow.id, workflow_attributes, session)
                session.flush()
            newly_created = True
        else:
            # set mutable attributes. Moved here from the set_resume method
            workflow.description = description
            workflow.name = name
            workflow.max_concurrently_running = max_concurrently_running
            session.flush()

            # upsert attributes
            if workflow_attributes:
                logger.info("Upsert attributes for workflow")
                if workflow_attributes:
                    for name, val in workflow_attributes.items():
                        _upsert_wf_attribute(workflow.id, name, val, session)
            newly_created = False

    resp = jsonify(
        {
            "workflow_id": workflow.id,
            "status": workflow.status,
            "newly_created": newly_created,
        }
    )
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow/<workflow_args_hash>", methods=["GET"])
def get_matching_workflows_by_workflow_args(workflow_args_hash: str) -> Any:
    """Return any dag hashes that are assigned to workflows with identical workflow args."""
    try:
        workflow_args_hash = str(int(workflow_args_hash))
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    structlog.contextvars.bind_contextvars(workflow_args_hash=str(workflow_args_hash))
    logger.info(f"Looking for wf with hash {workflow_args_hash}")

    session = SessionLocal()
    with session.begin():
        select_stmt = (
            select(Workflow.task_hash, Workflow.tool_version_id, Dag.hash)
            .join_from(Workflow, Dag, Workflow.dag_id == Dag.id)
            .where(Workflow.workflow_args_hash == workflow_args_hash)
        )
        res = []
        for row in session.execute(select_stmt).all():
            res.append((row.task_hash, row.tool_version_id, row.hash))

    if len(res) > 0:
        logger.debug(
            f"Found {res} workflow for " f"workflow_args_hash {workflow_args_hash}"
        )

    resp = jsonify(matching_workflows=res)
    resp.status_code = StatusCodes.OK
    return resp


def _add_or_get_wf_attribute_type(name: str, session: Session) -> int:
    try:
        with session.begin_nested():
            wf_attrib_type = WorkflowAttributeType(name=name)
            session.add(wf_attrib_type)
    except sqlalchemy.exc.IntegrityError:
        with session.begin_nested():
            select_stmt = select(WorkflowAttributeType).where(
                WorkflowAttributeType.name == name
            )
            wf_attrib_type = session.execute(select_stmt).scalars().one()

    return wf_attrib_type.id


def _upsert_wf_attribute(
    workflow_id: int, name: str, value: str, session: Session
) -> None:
    with session.begin_nested():
        wf_attrib_id = _add_or_get_wf_attribute_type(name, session)
        if SessionLocal.bind.dialect.name == "mysql":
            insert_vals = mysql_insert(WorkflowAttribute).values(
                workflow_id=workflow_id,
                workflow_attribute_type_id=wf_attrib_id,
                value=value,
            )
            upsert_stmt = insert_vals.on_duplicate_key_update(
                value=insert_vals.inserted.value
            )
        elif SessionLocal.bind.dialect.name == "sqlite":
            insert_vals = sqlite_insert(WorkflowAttribute).values(
                workflow_id=workflow_id,
                workflow_attribute_type_id=wf_attrib_id,
                value=value,
            )
            upsert_stmt = insert_vals.on_conflict_do_update(
                index_elements=["workflow_id", "workflow_attribute_type_id"],
                set_=dict(value=value),
            )
        session.execute(upsert_stmt)


@blueprint.route("/workflow/<workflow_id>/workflow_attributes", methods=["PUT"])
def update_workflow_attribute(workflow_id: int) -> Any:
    """Update the attributes for a given workflow."""
    structlog.contextvars.bind_contextvars(workflow_id=workflow_id)
    try:
        workflow_id = int(workflow_id)
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e
    """ Add/update attributes for a workflow """
    data = cast(Dict, request.get_json())

    logger.debug("Update attributes")
    attributes = data["workflow_attributes"]
    if attributes:
        session = SessionLocal()
        with session.begin():
            for name, val in attributes.items():
                _upsert_wf_attribute(workflow_id, name, val, session)

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow/<workflow_id>/set_resume", methods=["POST"])
def set_resume(workflow_id: int) -> Any:
    """Set resume on a workflow."""
    structlog.contextvars.bind_contextvars(workflow_id=workflow_id)
    try:
        data = cast(Dict, request.get_json())
        logger.info("Set resume for workflow")
        reset_running_jobs = bool(data["reset_running_jobs"])
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    session = SessionLocal()
    with session.begin():
        select_stmt = select(Workflow).where(Workflow.id == workflow_id)
        workflow = session.execute(select_stmt).scalars().one_or_none()
        if workflow:
            # trigger resume on active workflow run
            workflow.resume(reset_running_jobs)
            session.flush()
            logger.info(f"Resume set for wf {workflow_id}")

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow/<workflow_id>/is_resumable", methods=["GET"])
def workflow_is_resumable(workflow_id: int) -> Any:
    """Check if a workflow is in a resumable state."""
    structlog.contextvars.bind_contextvars(workflow_id=workflow_id)

    session = SessionLocal()
    with session.begin():
        select_stmt = select(Workflow).where(Workflow.id == workflow_id)
        workflow = session.execute(select_stmt).scalars().one()

    logger.info(f"Workflow is resumable: {workflow.is_resumable}")
    resp = jsonify(workflow_is_resumable=workflow.is_resumable)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/workflow/<workflow_id>/get_max_concurrently_running", methods=["GET"]
)
def get_max_concurrently_running(workflow_id: int) -> Any:
    """Return the maximum concurrency of this workflow."""
    structlog.contextvars.bind_contextvars(workflow_id=workflow_id)

    session = SessionLocal()
    with session.begin():
        select_stmt = select(Workflow).where(Workflow.id == workflow_id)
        workflow = session.execute(select_stmt).scalars().one()

    resp = jsonify(max_concurrently_running=workflow.max_concurrently_running)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "workflow/<workflow_id>/update_max_concurrently_running", methods=["PUT"]
)
def update_max_running(workflow_id: int) -> Any:
    """Update the number of tasks that can be running concurrently for a given workflow."""
    data = cast(Dict, request.get_json())
    structlog.contextvars.bind_contextvars(workflow_id=workflow_id)
    logger.debug("Update workflow max concurrently running")

    try:
        new_limit = data["max_tasks"]
    except KeyError as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    session = SessionLocal()
    with session.begin():
        update_stmt = (
            update(Workflow)
            .where(Workflow.id == workflow_id)
            .values(max_concurrently_running=new_limit)
        )
        res = session.execute(update_stmt)

    if res.rowcount == 0:  # Return a warning message if no update was performed
        message = (
            f"No update performed for workflow ID {workflow_id}, max_concurrently_running is "
            f"{new_limit}"
        )
    else:
        message = (
            f"Workflow ID {workflow_id} max concurrently running updated to {new_limit}"
        )

    resp = jsonify(message=message)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow/<workflow_id>/task_status_updates", methods=["POST"])
def task_status_updates(workflow_id: int) -> Any:
    """Returns all tasks in the database that have the specified status.

    Args:
        workflow_id (int): the ID of the workflow.
    """
    structlog.contextvars.bind_contextvars(workflow_id=workflow_id)
    data = cast(Dict, request.get_json())
    logger.info("Get task by status")

    try:
        filter_criteria: Tuple = (
            (Task.workflow_id == workflow_id),
            (Task.status_date >= data["last_sync"]),
        )
    except KeyError:
        filter_criteria = (Task.workflow_id == workflow_id,)

    # get time from db
    session = SessionLocal()
    with session.begin():
        db_time = session.execute(select(func.now())).scalar()
        str_time = db_time.strftime("%Y-%m-%d %H:%M:%S")

        tasks_by_status_query = (
            select(Task.status, func.group_concat(Task.id))
            .where(*filter_criteria)
            .group_by(Task.status)
        )
        result_dict = {}
        for row in session.execute(tasks_by_status_query):
            result_dict[row[0]] = [int(i) for i in row[1].split(",")]

    resp = jsonify(tasks_by_status=result_dict, time=str_time)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow/<workflow_id>/fetch_workflow_metadata", methods=["GET"])
def fetch_workflow_metadata(workflow_id: int) -> Any:
    """Get metadata associated with specified Workflow ID."""
    # Query for a workflow object
    session = SessionLocal()
    with session.begin():
        wf = session.execute(
            select(Workflow).where(Workflow.id == workflow_id)
        ).scalar()

    if not wf:
        logger.warning(f"No workflow found for ID {workflow_id}")
        return_tuple = ()
    else:
        return_tuple = wf.to_wire_as_distributor_workflow()

    resp = jsonify(workflow=return_tuple)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow/get_tasks/<workflow_id>", methods=["GET"])
def get_tasks_from_workflow(workflow_id: int) -> Any:
    """Return tasks associated with specified Workflow ID."""
    max_task_id = request.args.get("max_task_id")
    chunk_size = request.args.get("chunk_size")

    session = SessionLocal()

    with session.begin():
        # Query task table
        query = (
            select(
                Task.id,
                Task.array_id,
                Task.status,
                Task.max_attempts,
                Task.resource_scales,
                Task.fallback_queues,
                TaskResources.requested_resources,
                TaskResources.queue_id,
            )
            .join_from(Task, Array, Task.array_id == Array.id)
            .join_from(Task, TaskResources, Task.task_resources_id == TaskResources.id)
            .where(
                Task.workflow_id == workflow_id,
                # Note: because of this status != "DONE" filter, only the portion of the DAG
                # that is not complete is returned. Assumes that all tasks in a workflow
                # correspond to nodes that belong in the same DAG, and that no downstream
                # nodes can be in DONE for any unfinished task
                Task.status != TaskStatus.DONE,
                # Greater than set by the input max_task_id
                Task.id > max_task_id,
            )
            .order_by(Task.id)
            .limit(chunk_size)
        )
        res = session.execute(query).all()

        queue_map: Dict[int, List[int]] = {}
        array_map: Dict[int, List[int]] = {}
        resp_dict = {}
        for row in res:
            task_id = row[0]
            array_id = row[1]
            queue_id = row[7]
            row_metadata = row[1:7]

            resp_dict[task_id] = list(row_metadata)
            if queue_id not in queue_map:
                queue_map[queue_id] = []
            queue_map[queue_id].append(task_id)
            if array_id not in array_map:
                array_map[array_id] = []
            array_map[array_id].append(task_id)

        # get the queue and cluster
        for queue_id in queue_map.keys():
            queue = session.get(Queue, queue_id)
            queue_name = queue.name
            cluster_name = queue.cluster.name
            for task_id in queue_map[queue_id]:
                resp_dict[task_id].extend([cluster_name, queue_name])

        # get the max concurrency
        for array_id in array_map.keys():
            array = session.get(Array, array_id)
            max_concurrently_running = array.max_concurrently_running
            for task_id in array_map[array_id]:
                resp_dict[task_id].append(max_concurrently_running)

    resp = jsonify(tasks=resp_dict)
    resp.status_code = StatusCodes.OK
    return resp



"""Array Table for the Database."""

from sqlalchemy import Column, DateTime, Integer, String, UniqueConstraint
from sqlalchemy.sql import func

from jobmon.core.serializers import SerializeDistributorArray
from jobmon.server.web.models import Base


class Array(Base):
    """Array Database object."""

    __tablename__ = "array"

    def to_wire_as_distributor_array(self) -> tuple:
        """Serialize executor task object."""
        serialized = SerializeDistributorArray.to_wire(
            array_id=self.id,
            max_concurrently_running=self.max_concurrently_running,
            name=self.name,
        )
        return serialized

    id = Column(Integer, primary_key=True)
    name = Column(String(255), index=True)
    task_template_version_id = Column(Integer, index=True)
    workflow_id = Column(Integer, index=True)
    max_concurrently_running = Column(Integer)
    created_date = Column(DateTime, default=func.now())

    __table_args__ = (
        UniqueConstraint(
            "task_template_version_id",
            "workflow_id",
            name="uc_task_template_version_id_workflow_id",
        ),
    )



"""Task Instance Database Table."""
import json
from typing import Tuple

from sqlalchemy import Column, DateTime, ForeignKey, Index, Integer, String, Text
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func
import structlog

from jobmon.core.exceptions import InvalidStateTransition
from jobmon.core.serializers import SerializeTaskInstance
from jobmon.server.web.models import Base
from jobmon.server.web.models.task_instance_status import TaskInstanceStatus
from jobmon.server.web.models.task_status import TaskStatus


# new structlog logger per flask request context. internally stored as flask.g.logger
logger = structlog.get_logger(__name__)


class TaskInstance(Base):
    """Task Instance Database Table."""

    __tablename__ = "task_instance"

    def to_wire_as_distributor_task_instance(self) -> Tuple:
        """Serialize task instance object."""
        return SerializeTaskInstance.to_wire_distributor(
            self.id,
            self.task_id,
            self.workflow_run_id,
            self.task.workflow_id,
            self.status,
            self.distributor_id,
            self.cluster_id,
            self.task_resources_id,
            self.array_id,
            self.array_batch_num,
            self.array_step_id,
        )

    def to_wire_as_worker_node_task_instance(self) -> Tuple:
        """Serialize task instance object."""
        requested_resources = json.loads(self.task_resources.requested_resources)
        return SerializeTaskInstance.to_wire_worker_node(
            self.id,
            self.status,
            self.workflow_run_id,
            self.task.id,
            self.task.name,
            self.task.command,
            self.task.workflow_id,
            requested_resources.get("stdout"),
            requested_resources.get("stderr"),
        )

    id = Column(Integer, primary_key=True)
    workflow_run_id = Column(Integer, ForeignKey("workflow_run.id"))
    array_id = Column(Integer, ForeignKey("array.id"), default=None)
    task_id = Column(Integer, ForeignKey("task.id"))
    task_resources_id = Column(Integer, ForeignKey("task_resources.id"), index=True)
    array_batch_num = Column(Integer, index=True)
    array_step_id = Column(Integer, index=True)

    distributor_id = Column(String(20), index=True)

    # usage
    nodename = Column(String(150))
    process_group_id = Column(Integer)
    usage_str = Column(String(250))
    wallclock = Column(String(50))
    maxrss = Column(String(50))
    maxpss = Column(String(50))
    cpu = Column(String(50))
    io = Column(String(50))
    stdout = Column(String(2048))
    stderr = Column(String(2048))
    stdout_log = Column(Text)
    stderr_log = Column(Text)

    # status/state
    status = Column(
        String(1),
        ForeignKey("task_instance_status.id"),
        default=TaskInstanceStatus.QUEUED,
    )
    submitted_date = Column(DateTime)
    status_date = Column(DateTime, default=func.now())
    report_by_date = Column(DateTime)

    # ORM relationships
    task = relationship("Task", back_populates="task_instances")
    errors = relationship("TaskInstanceErrorLog", back_populates="task_instance")
    task_resources = relationship("TaskResources")

    __table_args__ = (
        Index(
            "ix_array_batch_index",
            "array_id",
            "array_batch_num",
            "array_step_id",
        ),
        Index("ix_status_status_date", "status", "status_date"),
    )

    # finite state machine transition information
    valid_transitions = [
        # task instance is moved from queued to instantiated by distributor
        (TaskInstanceStatus.QUEUED, TaskInstanceStatus.INSTANTIATED),
        # task instance is queued and waiting to instantiate when a new workflow run starts and
        # tells it to die
        (TaskInstanceStatus.QUEUED, TaskInstanceStatus.KILL_SELF),
        # task instance is launched by distributor
        (TaskInstanceStatus.INSTANTIATED, TaskInstanceStatus.LAUNCHED),
        # task instance submission hit weird bug and didn't get an distributor_id
        (TaskInstanceStatus.INSTANTIATED, TaskInstanceStatus.NO_DISTRIBUTOR_ID),
        # task instance is mid submission and a new workflow run starts and
        # tells it to die
        (TaskInstanceStatus.INSTANTIATED, TaskInstanceStatus.KILL_SELF),
        # task instance logs running before submitted due to race condition
        (TaskInstanceStatus.INSTANTIATED, TaskInstanceStatus.RUNNING),
        # task instance running after transitioning from launched
        (TaskInstanceStatus.LAUNCHED, TaskInstanceStatus.RUNNING),
        # task instance disappeared from distributor heartbeat and never logged
        # running. The distributor has no accounting of why it died
        (TaskInstanceStatus.LAUNCHED, TaskInstanceStatus.UNKNOWN_ERROR),
        # task instance disappeared from distributor heartbeat and never logged
        # running. The distributor discovered a resource error exit status.
        # This seems unlikely but is valid for the purposes of the FSM
        (TaskInstanceStatus.LAUNCHED, TaskInstanceStatus.RESOURCE_ERROR),
        # task instance is submitted to the batch distributor waiting to launch.
        # new workflow run is created and this task is told to kill
        # itself
        (TaskInstanceStatus.LAUNCHED, TaskInstanceStatus.KILL_SELF),
        # allow task instance to transit to F to immediately fail the task if there is an env
        # mismatch
        (TaskInstanceStatus.LAUNCHED, TaskInstanceStatus.ERROR_FATAL),
        # task instance triaging after transitioning from running
        (TaskInstanceStatus.RUNNING, TaskInstanceStatus.TRIAGING),
        # task instance hits an application error (happy path)
        (TaskInstanceStatus.RUNNING, TaskInstanceStatus.ERROR),
        # task instance stops logging heartbeats. reconciler can't find an exit
        # status
        (TaskInstanceStatus.RUNNING, TaskInstanceStatus.UNKNOWN_ERROR),
        # 1) task instance stops logging heartbeats. reconciler discovers a
        # resource error.
        # 2) worker node detects a resource error
        (TaskInstanceStatus.RUNNING, TaskInstanceStatus.RESOURCE_ERROR),
        # task instance is running. another workflow run starts and tells it to
        # die
        (TaskInstanceStatus.RUNNING, TaskInstanceStatus.KILL_SELF),
        # task instance finishes normally (happy path)
        (TaskInstanceStatus.RUNNING, TaskInstanceStatus.DONE),
        # task instance launched after transitioning from triaging
        (TaskInstanceStatus.TRIAGING, TaskInstanceStatus.RUNNING),
        # task instance resource_error after transitioning from triaging
        (TaskInstanceStatus.TRIAGING, TaskInstanceStatus.RESOURCE_ERROR),
        # task instance unknown_error after transitioning from triaging
        (TaskInstanceStatus.TRIAGING, TaskInstanceStatus.UNKNOWN_ERROR),
        # task instance error_fatal after transitioning from triaging
        (TaskInstanceStatus.TRIAGING, TaskInstanceStatus.ERROR_FATAL),
        # task instance error after transitioning from kill_self
        (TaskInstanceStatus.KILL_SELF, TaskInstanceStatus.ERROR_FATAL),
    ]

    untimely_transitions = [
        # task instance logs running before the distributor logs submitted due to
        # race condition. this is unlikely but happens and is valid for the
        # purposes of the FSM
        (TaskInstanceStatus.RUNNING, TaskInstanceStatus.LAUNCHED),
        # A worker node instance can error before the distributor moves it to launched.
        (TaskInstanceStatus.ERROR, TaskInstanceStatus.LAUNCHED),
        # task instance stops logging heartbeats and reconciler is looking for
        # remote exit status but can't find it so logs an unknown error. task
        # finishes with an application error. We can't update state because
        # the task may already be running again due to a race with the JIF
        (TaskInstanceStatus.ERROR, TaskInstanceStatus.UNKNOWN_ERROR),
        # task instance stops logging heartbeats and reconciler can't find exit
        # status. Worker tries to finish gracefully but reconciler won the race
        (TaskInstanceStatus.UNKNOWN_ERROR, TaskInstanceStatus.DONE),
        # task instance stops logging heartbeats and reconciler can't find exit
        # status. Worker tries to report an application error but cant' because
        # the task could be running again alread and we don't want to update
        # task state
        (TaskInstanceStatus.UNKNOWN_ERROR, TaskInstanceStatus.ERROR),
        # task instance stops logging heartbeats and reconciler can't find exit
        # status. Worker tries to report a resource error but cant' because
        # the task could be running again alread and we don't want to update
        # task state
        (TaskInstanceStatus.UNKNOWN_ERROR, TaskInstanceStatus.RESOURCE_ERROR),
        # task instance stops logging heartbeats and reconciler can't find exit
        # status. Worker reports a resource error before reconciler logs an
        # unknown error.
        (TaskInstanceStatus.RESOURCE_ERROR, TaskInstanceStatus.UNKNOWN_ERROR),
        # task instance stops logging heartbeats and reconciler is looking for
        # remote exit status but can't find it so logs an unknown error.
        # The worker finishes gracefully before reconciler can log an unknown
        # error
        (TaskInstanceStatus.DONE, TaskInstanceStatus.UNKNOWN_ERROR),
        # task is reset by workflow resume and worker finishes gracefully but
        # resume won the race
        (TaskInstanceStatus.KILL_SELF, TaskInstanceStatus.DONE),
        # task is reset by workflow resume and reconciler or worker node
        # discovers resource error, but resume won the race
        (TaskInstanceStatus.KILL_SELF, TaskInstanceStatus.RESOURCE_ERROR),
    ]

    error_states = [
        TaskInstanceStatus.NO_DISTRIBUTOR_ID,
        TaskInstanceStatus.ERROR,
        TaskInstanceStatus.UNKNOWN_ERROR,
        TaskInstanceStatus.RESOURCE_ERROR,
        TaskInstanceStatus.KILL_SELF,
    ]

    def transition(self, new_state: str) -> None:
        """Transition the TaskInstance status."""
        # if the transition is timely, move to new state. Otherwise do nothing
        # bind_to_logger(
        #     workflow_run_id=self.workflow_run_id,
        #     task_id=self.task_id,
        #     task_instance_id=self.id,
        # )
        if self._is_timely_transition(new_state):
            self._validate_transition(new_state)
            logger.info(
                f"Transitioning task_instance from {self.status} to {new_state}"
            )
            self.status = new_state
            self.status_date = func.now()
            if new_state == TaskInstanceStatus.QUEUED:
                self.task.transition(TaskStatus.QUEUED)
            if new_state == TaskInstanceStatus.INSTANTIATED:
                self.task.transition(TaskStatus.INSTANTIATING)
            if new_state == TaskInstanceStatus.LAUNCHED:
                self.task.transition(TaskStatus.LAUNCHED)
            if new_state == TaskInstanceStatus.RUNNING:
                self.task.transition(TaskStatus.RUNNING)
            elif new_state == TaskInstanceStatus.DONE:
                self.task.transition(TaskStatus.DONE)
            elif new_state in self.error_states:
                self.task.transition_after_task_instance_error(new_state)
            elif new_state == TaskInstanceStatus.ERROR_FATAL:
                # if the task instance is F, the task status should be F too
                self.task.transition(TaskStatus.ERROR_RECOVERABLE)
                self.task.transition(TaskStatus.ERROR_FATAL)

    def _validate_transition(self, new_state: str) -> None:
        """Ensure the TaskInstance status transition is valid."""
        if (self.status, new_state) not in self.__class__.valid_transitions:
            raise InvalidStateTransition(
                "TaskInstance", self.id, self.status, new_state
            )

    def _is_timely_transition(self, new_state: str) -> bool:
        """Check if the transition is invalid due to a race condition."""
        if (self.status, new_state) in self.__class__.untimely_transitions:
            msg = str(
                InvalidStateTransition("TaskInstance", self.id, self.status, new_state)
            )
            msg += (
                ". This is an untimely transition likely caused by a race "
                " condition between the distributor_service and the worker_node."
            )
            logger.warning(msg)
            return False
        else:
            return True



"""Task Instance Status Table."""
from sqlalchemy import Column, String
from sqlalchemy.orm import Session

from jobmon.core.constants import TaskInstanceStatus as Statuses
from jobmon.server.web.models import Base


class TaskInstanceStatus(Base):
    """The table in the database that holds on the possible statuses for TaskInstance."""

    __tablename__ = "task_instance_status"

    QUEUED = Statuses.QUEUED
    INSTANTIATED = Statuses.INSTANTIATED
    NO_DISTRIBUTOR_ID = Statuses.NO_DISTRIBUTOR_ID
    LAUNCHED = Statuses.LAUNCHED
    RUNNING = Statuses.RUNNING
    TRIAGING = Statuses.TRIAGING
    RESOURCE_ERROR = Statuses.RESOURCE_ERROR
    UNKNOWN_ERROR = Statuses.UNKNOWN_ERROR
    ERROR = Statuses.ERROR
    DONE = Statuses.DONE
    KILL_SELF = Statuses.KILL_SELF
    ERROR_FATAL = Statuses.ERROR_FATAL

    id = Column(String(1), primary_key=True)
    label = Column(String(150))
    description = Column(String(150))


def add_task_instance_statuses(session: Session) -> None:
    """Populate the task_instance_status table in the database, in alphabetical order."""
    statuses = [
        TaskInstanceStatus(
            id="D", label="DONE", description="Task instance finished successfully."
        ),
        TaskInstanceStatus(
            id="E",
            label="ERROR",
            description="Task instance stopped with an application error "
            "(non-zero return code).",
        ),
        TaskInstanceStatus(
            id="F",
            label="ERROR_FATAL",
            description="Task instance killed itself as part of a cold workflow resume, "
            "and cannot be retried.",
        ),
        TaskInstanceStatus(
            id="I",
            label="INSTANTIATED",
            description="Task instance is created within Jobmon, but not queued for "
            "submission to the cluster.",
        ),
        TaskInstanceStatus(
            id="K",
            label="KILL_SELF",
            description="Task instance has been ordered to kill itself if it is still alive, "
            "as part of a cold workflow resume.",
        ),
        TaskInstanceStatus(
            id="O",
            label="QUEUED",
            description="Task instance submitted to the cluster normally, "
            "part of a Job Array.",
        ),
        TaskInstanceStatus(
            id="Q",
            label="QUEUED",
            description="TaskInstance is queued for submission to the cluster.",
        ),
        TaskInstanceStatus(
            id="R",
            label="RUNNING",
            description="Task instance has started running normally.",
        ),
        TaskInstanceStatus(
            id="T",
            label="TRIAGING",
            description="Task instance has errored, Jobmon "
            "is determining the category of error.",
        ),
        TaskInstanceStatus(
            id="U",
            label="UNKNOWN_ERROR",
            description="Task instance stopped reporting that it was alive "
            "for an unknown reason.",
        ),
        TaskInstanceStatus(
            id="W",
            label="NO_DISTRIBUTOR_ID",
            description="Task instance submission within Jobmon failed – "
            "did not receive a distributor_id from the cluster.",
        ),
        TaskInstanceStatus(
            id="Z",
            label="RESOURCE_ERROR",
            description="Task instance died because of insufficient resource request, "
            "i.e. insufficient memory or runtime.",
        ),
    ]
    session.add_all(statuses)



"""Add handlers to deal with server-side exceptions and logging."""
from typing import Any, cast, Dict, Optional

from elasticapm.contrib.flask import ElasticAPM
from flask import Flask, jsonify, request
from MySQLdb import OperationalError
import structlog
from werkzeug.exceptions import BadRequest


from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.server_side_exception import InvalidUsage, ServerError


# new structlog logger per flask request context. internally stored as flask.g.logger
logger = structlog.get_logger(__name__)


def add_hooks_and_handlers(app: Flask, apm: Optional[ElasticAPM] = None) -> Flask:
    """Add logging hooks and exception handlers."""

    @app.errorhandler(Exception)
    def handle_anything(error: Any) -> Any:
        if apm is not None:
            apm.capture_exception(exc_info=(type(error), error, error.__traceback__))
        try:
            status_code = error.status_code
        except AttributeError:
            status_code = 500

        if "deadlock found" in str(error).lower():
            status_code = 423

        response_dict = {
            "type": str(type(error)),
            "exception_message": str(error),
            "status_code": str(status_code),
        }
        logger.exception(
            "server encountered:", status_code=status_code, route=request.path
        )
        response = jsonify(error=response_dict)
        response.content_type = "application/json"
        response.status_code = status_code
        return response

    # handle 404 at the application level not the blueprint level
    @app.errorhandler(404)
    def page_not_found(e: Any) -> tuple:
        return f"This route does not exist: {request.url}", 404

    # error handling
    @app.errorhandler(InvalidUsage)
    def handle_4xx(error: InvalidUsage) -> Any:
        logger.exception(
            "server encountered:", status_code=error.status_code, route=request.path
        )
        if apm is not None:
            apm.capture_exception(exc_info=(type(error), error, error.__traceback__))
        response_dict = {"type": str(type(error)), "exception_message": str(error)}
        response = jsonify(error=response_dict)
        response.content_type = "application/json"
        response.status_code = error.status_code
        return response

    # error handling
    @app.errorhandler(ServerError)
    def handle_5xx(error: ServerError) -> Any:
        logger.exception(
            "server encountered:", status_code=error.status_code, route=request.path
        )
        if apm is not None:
            apm.capture_exception(exc_info=(type(error), error, error.__traceback__))
        response_dict = {"type": str(type(error)), "exception_message": str(error)}
        response = jsonify(error=response_dict)
        response.content_type = "application/json"
        response.status_code = error.status_code
        return response

    # error handling
    @app.errorhandler(OperationalError)
    def handle_mysql_gone_away(error: OperationalError) -> Any:
        if "2013, 'Lost connection to MySQL server during query'" in str(error):
            engine = SessionLocal().get_bind()
            # A new connection pool is created immediately after the old one has been disposed
            engine.dispose()

    @app.before_request
    def add_requester_context() -> None:
        # get a clean threadlocal structlog context for this request
        structlog.contextvars.clear_contextvars()

        server_structlog_context = None

        try:
            data = cast(Dict, request.get_json())
        except BadRequest:
            # Some get requests come without any json data.
            # All requests issued by Jobmon's requester automatically come with an empty dict;
            # however, if we call raw get requests outside of Jobmon we should handle it
            data = {}
        if request.method == "GET":
            server_structlog_context = data
        if request.method in ["POST", "PUT"]:
            server_structlog_context = data.pop("server_structlog_context", {})
        if server_structlog_context:
            structlog.contextvars.bind_contextvars(
                path=request.path, **server_structlog_context
            )

    return app



"""Web API subpackage."""
from sqlalchemy import orm

from jobmon.server.web.log_config import configure_structlog

configure_structlog()

# configurable session factory. add an engine using session_factory.configure(bind=eng)
session_factory = orm.sessionmaker(future=True)



"""Workflow run database table."""
import datetime

from sqlalchemy import Column, DateTime, ForeignKey, Index, Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func
import structlog

from jobmon.core.exceptions import InvalidStateTransition
from jobmon.core.serializers import SerializeWorkflowRun
from jobmon.server import __version__
from jobmon.server.web._compat import add_time
from jobmon.server.web.models import Base
from jobmon.server.web.models.workflow_run_status import WorkflowRunStatus
from jobmon.server.web.models.workflow_status import WorkflowStatus


logger = structlog.get_logger(__name__)


class WorkflowRun(Base):
    """Database table for recording Workflow Runs."""

    __tablename__ = "workflow_run"

    def to_wire_as_reaper_workflow_run(self) -> tuple:
        """Serialize workflow run."""
        serialized = SerializeWorkflowRun.to_wire(
            id=self.id, workflow_id=self.workflow_id
        )
        return serialized

    id = Column(Integer, primary_key=True)
    workflow_id = Column(Integer, ForeignKey("workflow.id"))
    user = Column(String(150))
    jobmon_version = Column(String(150), default="UNKNOWN")
    jobmon_server_version = Column(String(150), default=__version__)
    status = Column(
        String(1),
        ForeignKey("workflow_run_status.id"),
        default=WorkflowRunStatus.REGISTERED,
    )

    created_date = Column(DateTime, default=func.now())
    status_date = Column(DateTime, default=func.now())
    heartbeat_date = Column(DateTime, default=func.now())

    workflow = relationship("Workflow", back_populates="workflow_runs", lazy=True)

    __table_args__ = (
        Index("ix_status_version", "status", "jobmon_version", "jobmon_server_version"),
    )

    valid_transitions = [
        # a workflow run is created normally. claimed control of workflow
        (WorkflowRunStatus.REGISTERED, WorkflowRunStatus.LINKING),
        # a workflow run is created normally. All tasks are updated in the db
        # and the workflow run can move to bound state
        (WorkflowRunStatus.LINKING, WorkflowRunStatus.BOUND),
        # a workflow run is created normally. Something goes wrong while the
        # tasks are binding and the workflow run moves to error state
        (WorkflowRunStatus.LINKING, WorkflowRunStatus.ABORTED),
        # a workflow run is bound and then moves to instantiating
        (WorkflowRunStatus.BOUND, WorkflowRunStatus.INSTANTIATED),
        # a workflow run moves from instantiating to launched
        (WorkflowRunStatus.INSTANTIATED, WorkflowRunStatus.LAUNCHED),
        # a workflow run moves from launched to running
        (WorkflowRunStatus.LAUNCHED, WorkflowRunStatus.RUNNING),
        # a workflow run can't be launched for some reason. TODO: implement triaging
        (WorkflowRunStatus.INSTANTIATED, WorkflowRunStatus.ERROR),
        (WorkflowRunStatus.LAUNCHED, WorkflowRunStatus.ERROR),
        # a workflow run is bound and then an error occurs before it starts
        # running
        (WorkflowRunStatus.BOUND, WorkflowRunStatus.ERROR),
        # a workflow run is bound and then a new workflow run is created
        # before the old workflow run moves into running state
        (WorkflowRunStatus.BOUND, WorkflowRunStatus.COLD_RESUME),
        (WorkflowRunStatus.BOUND, WorkflowRunStatus.HOT_RESUME),
        # the workflow starts running normally and finishes successfully
        (WorkflowRunStatus.RUNNING, WorkflowRunStatus.DONE),
        # the workflow starts running normally and the user stops it via a
        # keyboard interrupt
        (WorkflowRunStatus.RUNNING, WorkflowRunStatus.STOPPED),
        # the workflow is running and then a new workflow run is created
        (WorkflowRunStatus.RUNNING, WorkflowRunStatus.COLD_RESUME),
        (WorkflowRunStatus.RUNNING, WorkflowRunStatus.HOT_RESUME),
        # the workflow is running and then it's tasks hit errors
        (WorkflowRunStatus.RUNNING, WorkflowRunStatus.ERROR),
        # the workflow is set to resume and then it successfully shuts down
        (WorkflowRunStatus.COLD_RESUME, WorkflowRunStatus.TERMINATED),
        (WorkflowRunStatus.HOT_RESUME, WorkflowRunStatus.TERMINATED),
    ]

    untimely_transitions = [
        (WorkflowRunStatus.RUNNING, WorkflowRunStatus.RUNNING),
        (WorkflowRunStatus.LINKING, WorkflowRunStatus.LINKING),
    ]

    bound_error_states = [WorkflowRunStatus.STOPPED, WorkflowRunStatus.ERROR]

    active_states = [
        WorkflowRunStatus.BOUND,
        WorkflowRunStatus.RUNNING,
        WorkflowRunStatus.COLD_RESUME,
        WorkflowRunStatus.HOT_RESUME,
    ]

    @property
    def is_alive(self) -> bool:
        """Workflow run is in a state that should be registering heartbeats."""
        return self.status in [
            WorkflowRunStatus.LINKING,
            WorkflowRunStatus.BOUND,
            WorkflowRunStatus.RUNNING,
            WorkflowRunStatus.COLD_RESUME,
            WorkflowRunStatus.HOT_RESUME,
        ]

    @property
    def is_active(self) -> bool:
        """Statuses where Workflow Run is active (bound or running)."""
        return self.status in [WorkflowRunStatus.BOUND, WorkflowRunStatus.RUNNING]

    def terminable(self, current_time: datetime.datetime) -> bool:
        """Whether a workflowrun can be terminated.

        A workflowrun can be terminated if it is in Cold/Hot resume state and has missed
        the last reporting heartbeat.
        """
        return (
            (self.heartbeat_date <= current_time) and
            (self.status in (WorkflowRunStatus.COLD_RESUME, WorkflowRunStatus.HOT_RESUME))
        )

    def heartbeat(
        self,
        next_report_increment: float,
        transition_status: str = WorkflowRunStatus.RUNNING,
    ) -> None:
        """Register a heartbeat for the Workflow Run to show it is still alive."""
        self.transition(transition_status)
        self.heartbeat_date = add_time(next_report_increment)

    def reap(self) -> None:
        """Transition dead workflow runs to a terminal state."""
        structlog.contextvars.bind_contextvars(
            workflow_run_id=self.id, workflow_id=self.workflow_id
        )
        logger.info("Dead workflow_run will be reaped.")
        if self.status == WorkflowRunStatus.LINKING:
            logger.debug(f"Transitioning wfr {self.id} to ABORTED")
            self.transition(WorkflowRunStatus.ABORTED)
        if self.status in [WorkflowRunStatus.COLD_RESUME, WorkflowRunStatus.HOT_RESUME]:
            logger.debug(f"Transitioning wfr {self.id} to TERMINATED")
            self.transition(WorkflowRunStatus.TERMINATED)
        if self.status == WorkflowRunStatus.RUNNING:
            logger.debug(f"Transitioning wfr {self.id} to ERROR")
            self.transition(WorkflowRunStatus.ERROR)
        logger.info(f"Transitioned workflow to {self.status}")

    def transition(self, new_state: str) -> None:
        """Transition the Workflow Run's state."""
        structlog.contextvars.bind_contextvars(
            workflow_run_id=self.id, workflow_id=self.workflow_id
        )
        logger.info(f"Transitioning workflow_run from {self.status} to {new_state}")
        if self._is_timely_transition(new_state):
            self._validate_transition(new_state)
            self.status = new_state
            self.status_date = func.now()
            if new_state == WorkflowRunStatus.LINKING:
                self.workflow.transition(WorkflowStatus.REGISTERING)
            elif new_state == WorkflowRunStatus.BOUND:
                self.workflow.transition(WorkflowStatus.QUEUED)
            elif new_state == WorkflowRunStatus.ABORTED:
                self.workflow.transition(WorkflowStatus.ABORTED)
            elif new_state == WorkflowRunStatus.RUNNING:
                self.workflow.transition(WorkflowStatus.RUNNING)
            elif new_state == WorkflowRunStatus.DONE:
                self.workflow.transition(WorkflowStatus.DONE)
            elif new_state == WorkflowRunStatus.TERMINATED:
                self.workflow.transition(WorkflowStatus.HALTED)
            elif new_state in self.bound_error_states:
                self.workflow.transition(WorkflowStatus.FAILED)
            elif new_state == WorkflowRunStatus.INSTANTIATED:
                self.workflow.transition(WorkflowStatus.INSTANTIATING)
            elif new_state == WorkflowRunStatus.LAUNCHED:
                self.workflow.transition(WorkflowStatus.LAUNCHED)

    def hot_reset(self) -> None:
        """Set Workflow Run to Hot Resume."""
        structlog.contextvars.bind_contextvars(
            workflow_run_id=self.id, workflow_id=self.workflow_id
        )
        logger.info("Transitioning workflow_run to HOT_RESUME.")
        self.transition(WorkflowRunStatus.HOT_RESUME)

    def cold_reset(self) -> None:
        """Set Workflow Run to Cold Resume."""
        structlog.contextvars.bind_contextvars(
            workflow_run_id=self.id, workflow_id=self.workflow_id
        )
        logger.info("Transitioning workflow_run to COLD_RESUME.")
        self.transition(WorkflowRunStatus.COLD_RESUME)

    def _validate_transition(self, new_state: str) -> None:
        """Ensure the Job state transition is valid."""
        if (self.status, new_state) not in self.valid_transitions:
            raise InvalidStateTransition("WorkflowRun", self.id, self.status, new_state)

    def _is_timely_transition(self, new_state: str) -> bool:
        """Check if the transition is invalid due to a race condition."""
        structlog.contextvars.bind_contextvars(
            workflow_run_id=self.id, workflow_id=self.workflow_id
        )
        if (self.status, new_state) in self.untimely_transitions:
            logger.info(
                f"Ignoring transition of workflow_run from {self.status} to {new_state}"
            )
            return False
        else:
            logger.debug(
                f"No race condition when transitioning workflow_run from "
                f"{self.status} to {new_state}"
            )
            return True



"""Set up server specific CLI config."""
import argparse
import logging
import sys
from typing import Optional

from jobmon.core.cli import CLI

logger = logging.getLogger(__name__)


class ServerCLI(CLI):
    """CLI for Server only."""

    def __init__(self) -> None:
        """Initialize ServerCLI with subcommands."""
        self.parser = argparse.ArgumentParser("jobmon server")
        self._subparsers = self.parser.add_subparsers(dest="sub_command")

        # now add specific sub parsers
        self._add_workflow_reaper_subparser()
        self._add_init_db_subparser()
        self._add_terminate_db_subparser()

    def workflow_reaper(self, args: argparse.Namespace) -> None:
        """Workflow reaper entrypoint logic."""
        from jobmon.server.workflow_reaper.api import start_workflow_reaper

        logging.basicConfig(stream=sys.stdout, level=logging.INFO)
        if args.command == "start":
            start_workflow_reaper(
                service_url=args.service_url,
                slack_api_url=args.slack_api_url,
                slack_token=args.slack_token,
                slack_channel_default=args.slack_channel_default,
                poll_interval_minutes=args.poll_interval_minutes,
            )
        else:
            raise ValueError(
                "Invalid command choice. Options are (start), got " f"({args.command})"
            )

    def init_db(self, args: argparse.Namespace) -> None:
        """Entrypoint to initialize new Jobmon database."""
        import sqlalchemy
        from jobmon.core.configuration import JobmonConfig
        from jobmon.server.web.models import init_db

        sqlalchemy_database_uri = args.sqlalchemy_database_uri
        if not sqlalchemy_database_uri:
            config = JobmonConfig()
            sqlalchemy_database_uri = config.get("db", "sqlalchemy_database_uri")
        engine = sqlalchemy.create_engine(sqlalchemy_database_uri)
        init_db(engine)

    def terminate_db(self, args: argparse.Namespace) -> None:
        """Entrypoint to terminate a Jobmon database."""
        import sqlalchemy
        from jobmon.core.configuration import JobmonConfig
        from jobmon.server.web.models import terminate_db

        sqlalchemy_database_uri = args.sqlalchemy_database_uri
        if not sqlalchemy_database_uri:
            config = JobmonConfig()
            sqlalchemy_database_uri = config.get("db", "sqlalchemy_database_uri")
        engine = sqlalchemy.create_engine(sqlalchemy_database_uri)
        terminate_db(engine)

    def _add_workflow_reaper_subparser(self) -> None:
        reaper_parser = self._subparsers.add_parser("workflow_reaper")
        reaper_parser.set_defaults(func=self.workflow_reaper)
        reaper_parser.add_argument(
            "command",
            type=str,
            choices=["start"],
            help=(
                "The workflow_reaper sub-command to run: (start). Start command runs "
                "workflow_reaper.monitor_forever() method."
            ),
        )
        reaper_parser.add_argument(
            "--service_url",
            type=str,
            help="Jobmon web service URL",
            required=False,
            default="",
        )
        reaper_parser.add_argument(
            "--slack_api_url",
            type=str,
            help="URL to post notifications",
            required=False,
            default="",
        )
        reaper_parser.add_argument(
            "--slack_token",
            type=str,
            help="Authentication token for posting updates to slack",
            required=False,
            default="",
        )
        reaper_parser.add_argument(
            "--slack_channel_default",
            type=str,
            help="Default channel to post updates to",
            required=False,
            default="",
        )
        reaper_parser.add_argument(
            "--poll_interval_minutes",
            type=int,
            help="Duration in minutes to sleep between reaper loops",
            required=False,
            default=None,
        )

    def _add_init_db_subparser(self) -> None:
        init_db_parser = self._subparsers.add_parser("init_db")
        init_db_parser.set_defaults(func=self.init_db)
        init_db_parser.add_argument(
            "--sqlalchemy_database_uri",
            type=str,
            help="The connection string for sqlalchemy to use when running the server.",
            required=False,
            default="",
        )

    def _add_terminate_db_subparser(self) -> None:
        terminate_db_parser = self._subparsers.add_parser("terminate_db")
        terminate_db_parser.set_defaults(func=self.terminate_db)
        terminate_db_parser.add_argument(
            "--sqlalchemy_database_uri",
            type=str,
            help="The connection string for sqlalchemy to use when running the server.",
            required=False,
            default="",
        )


def main(argstr: Optional[str] = None) -> None:
    """Create CLI."""
    cli = ServerCLI()
    cli.main(argstr)



"""Workflow Run Status Database Table."""
from sqlalchemy import Column, String
from sqlalchemy.orm import Session

from jobmon.core.constants import WorkflowRunStatus as Statuses
from jobmon.server.web.models import Base


class WorkflowRunStatus(Base):
    """Workflow Run Status Database Table."""

    __tablename__ = "workflow_run_status"

    REGISTERED = Statuses.REGISTERED
    LINKING = Statuses.LINKING
    BOUND = Statuses.BOUND
    ABORTED = Statuses.ABORTED
    RUNNING = Statuses.RUNNING
    DONE = Statuses.DONE
    STOPPED = Statuses.STOPPED
    ERROR = Statuses.ERROR
    COLD_RESUME = Statuses.COLD_RESUME
    HOT_RESUME = Statuses.HOT_RESUME
    TERMINATED = Statuses.TERMINATED
    INSTANTIATED = Statuses.INSTANTIATED
    LAUNCHED = Statuses.LAUNCHED

    id = Column(String(1), primary_key=True)
    label = Column(String(150), nullable=False)
    description = Column(String(150))


def add_workflow_run_statuses(session: Session) -> None:
    """Populate the workflow_run_status table in the database."""
    statuses = [
        WorkflowRunStatus(
            id="A",
            label="ABORTED",
            description="WorkflowRun encountered problems while binding so it stopped.",
        ),
        WorkflowRunStatus(
            id="B",
            label="BOUND",
            description="WorkflowRun has been bound to the database.",
        ),
        WorkflowRunStatus(
            id="C",
            label="COLD_RESUME",
            description="WorkflowRun is set to resume as soon all existing tasks are killed.",
        ),
        WorkflowRunStatus(
            id="D",
            label="DONE",
            description="WorkflowRun is Done, it successfully completed.",
        ),
        WorkflowRunStatus(
            id="E",
            label="ERROR",
            description="WorkflowRun did not complete successfully, either some Tasks "
            "failed or (rarely) an internal Jobmon error.",
        ),
        WorkflowRunStatus(
            id="G", label="REGISTERED", description="WorkflowRun has been validated."
        ),
        WorkflowRunStatus(
            id="H",
            label="HOT_RESUME",
            description="WorkflowRun was set to hot-resume while tasks are still running, "
            "they will continue running.",
        ),
        WorkflowRunStatus(
            id="I",
            label="INSTANTIATED",
            description="Scheduler is instantiating a WorkflowRun on the distributor.",
        ),
        WorkflowRunStatus(
            id="L",
            label="LINKING",
            description="WorkflowRun completed successfully, updating the Workflow.",
        ),
        WorkflowRunStatus(
            id="O",
            label="LAUNCHED",
            description="Instantiation complete. Distributor is controlling Tasks or waiting "
            "for scheduling loop.",
        ),
        WorkflowRunStatus(
            id="R", label="RUNNING", description="WorkflowRun is currently running."
        ),
        WorkflowRunStatus(
            id="S",
            label="STOPPED",
            description="WorkflowRun was deliberately stopped, probably due to keyboard "
            "interrupt from user.",
        ),
        WorkflowRunStatus(
            id="T",
            label="TERMINATED",
            description="This WorkflowRun is being replaced by a new WorkflowRun created "
            "to pick up remaining Tasks, this WFR is terminating.",
        ),
    ]
    session.add_all(statuses)



"""Task Resources Type Database Table."""
from sqlalchemy import Column, String
from sqlalchemy.orm import Session

from jobmon.core.constants import TaskResourcesType as Types
from jobmon.server.web.models import Base


class TaskResourcesType(Base):
    """The table in the database that holds the possible statuses for the TaskResources."""

    __tablename__ = "task_resources_type"
    ORIGINAL = Types.ORIGINAL
    VALIDATED = Types.VALIDATED
    ADJUSTED = Types.ADJUSTED

    id = Column(String(1), primary_key=True)
    label = Column(String(150), nullable=False)


def add_task_resources_types(session: Session) -> None:
    """Populate the task_resources_type table in the database."""
    task_resources_type = [
        TaskResourcesType(id="O", label="ORIGINAL"),
        TaskResourcesType(id="V", label="VALIDATED"),
        TaskResourcesType(id="A", label="ADJUSTED"),
    ]
    session.add_all(task_resources_type)



"""Routes for Arrays."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict

from flask import jsonify, request
from sqlalchemy import bindparam, func, insert, literal_column, select, update
import structlog

from jobmon.core.constants import TaskInstanceStatus
from jobmon.server.web._compat import add_time
from jobmon.server.web.models.array import Array
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_instance import TaskInstance
from jobmon.server.web.models.task_status import TaskStatus
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint


logger = structlog.get_logger(__name__)


@blueprint.route("/array", methods=["POST"])
def add_array() -> Any:
    """Return an array ID by workflow and task template version ID.

    If not found, bind the array.
    """
    data = cast(Dict, request.get_json())
    workflow_id = int(data["workflow_id"])
    task_template_version_id = int(data["task_template_version_id"])

    structlog.contextvars.bind_contextvars(
        task_template_version_id=task_template_version_id,
        workflow_id=workflow_id,
    )

    # Check if the array is already bound, if so return it
    session = SessionLocal()
    with session.begin():
        select_stmt = select(Array).where(
            Array.workflow_id == workflow_id,
            Array.task_template_version_id == task_template_version_id,
        )
        array = session.execute(select_stmt).scalars().one_or_none()

        if array is None:  # not found, so need to add it
            array = Array(
                task_template_version_id=data["task_template_version_id"],
                workflow_id=data["workflow_id"],
                max_concurrently_running=data["max_concurrently_running"],
                name=data["name"],
            )
            session.add(array)
        else:
            update_stmt = (
                update(Array)
                .where(Array.id == array.id)
                .values(max_concurrently_running=data["max_concurrently_running"])
            )
            session.execute(update_stmt)
        session.commit()

    # return result
    resp = jsonify(array_id=array.id)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/array/<array_id>/queue_task_batch", methods=["POST"])
def record_array_batch_num(array_id: int) -> Any:
    """Record a batch number to associate sets of task instances with an array submission."""
    data = cast(Dict, request.get_json())
    array_id = int(array_id)
    task_ids = [int(task_id) for task_id in data["task_ids"]]
    task_resources_id = int(data["task_resources_id"])
    workflow_run_id = int(data["workflow_run_id"])

    session = SessionLocal()
    with session.begin():
        # update task status to acquire lock
        update_stmt = (
            update(Task)
            .where(
                Task.id.in_(task_ids),
                Task.status.in_(
                    [TaskStatus.REGISTERING, TaskStatus.ADJUSTING_RESOURCES]
                ),
            )
            .values(
                status=TaskStatus.QUEUED,
                status_date=func.now(),
                num_attempts=(Task.num_attempts + 1),
            )
        )
        session.execute(update_stmt)

        # now insert them into task instance
        insert_stmt = insert(TaskInstance).from_select(
            # columns map 1:1 to selected rows
            [
                "task_id",
                "workflow_run_id",
                "array_id",
                "task_resources_id",
                "array_batch_num",
                "array_step_id",
                "status",
                "status_date",
            ],
            # select statement
            select(
                # unique id
                Task.id.label("task_id"),
                # static associations
                literal_column(str(workflow_run_id)).label("workflow_run_id"),
                literal_column(str(array_id)).label("array_id"),
                literal_column(str(task_resources_id)).label("task_resources_id"),
                # batch info
                select(func.coalesce(func.max(TaskInstance.array_batch_num) + 1, 1))
                .where((TaskInstance.array_id == array_id))
                .label("array_batch_num"),
                (func.row_number().over(order_by=Task.id) - 1).label("array_step_id"),
                # status columns
                literal_column(f"'{TaskInstanceStatus.QUEUED}'").label("status"),
                func.now().label("status_date"),
            ).where(Task.id.in_(task_ids), Task.status == TaskStatus.QUEUED),
            # no python side defaults. Server defaults only
            include_defaults=False,
        )
        session.execute(insert_stmt)

    with session.begin():
        tasks_by_status_query = (
            select(Task.status, func.group_concat(Task.id))
            .where(Task.id.in_(task_ids))
            .group_by(Task.status)
        )
        result_dict = {}
        for row in session.execute(tasks_by_status_query):
            result_dict[row[0]] = [int(i) for i in row[1].split(",")]

    resp = jsonify(tasks_by_status=result_dict)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/array/<array_id>/transition_to_launched", methods=["POST"])
def transition_array_to_launched(array_id: int) -> Any:
    """Transition TIs associated with an array_id and batch_num to launched."""
    structlog.contextvars.bind_contextvars(array_id=array_id)

    data = cast(Dict, request.get_json())
    batch_num = data["batch_number"]
    next_report = data["next_report_increment"]

    session = SessionLocal()
    with session.begin():
        # Acquire a lock and update tasks to launched
        task_ids_query = (
            select(TaskInstance.task_id)
            .where(
                TaskInstance.array_id == array_id,
                TaskInstance.array_batch_num == batch_num,
            )
            .execution_options(synchronize_session=False)
        )

        task_ids = session.execute(task_ids_query).scalars()

        update_task_stmt = (
            update(Task)
            .where(
                Task.array_id == array_id,
                Task.id.in_(task_ids),
                Task.status == TaskStatus.INSTANTIATING,
            )
            .values(status=TaskStatus.LAUNCHED, status_date=func.now())
        ).execution_options(synchronize_session=False)
        session.execute(update_task_stmt)

        # Transition all the task instances in the batch
        # Bypassing the ORM for performance reasons.
        update_stmt = (
            update(TaskInstance)
            .where(
                TaskInstance.array_id == array_id,
                TaskInstance.status == TaskInstanceStatus.INSTANTIATED,
                TaskInstance.array_batch_num == batch_num,
            )
            .values(
                status=TaskInstanceStatus.LAUNCHED,
                submitted_date=func.now(),
                status_date=func.now(),
                report_by_date=add_time(next_report),
            )
        ).execution_options(synchronize_session=False)

        session.execute(update_stmt)

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/array/<array_id>/log_distributor_id", methods=["POST"])
def log_array_distributor_id(array_id: int) -> Any:
    """Add distributor_id, stderr/stdout paths to the DB for all TIs in an array."""
    data = cast(Dict, request.get_json())

    # Create a list of dicts out of the distributor id map.
    params = [
        {"task_instance_id": key, "distributor_id": val} for key, val in data.items()
    ]

    session = SessionLocal()
    with session.begin():
        # Acquire a lock and update the task instance table
        # Using bindparam only issues one query; unfortunately, the MariaDB optimizer actually
        # performs this operation iteratively. The update is fairly slow despite the fact that
        # we are issuing a single bulk query.
        update_stmt = (
            update(TaskInstance)
            .where(
                TaskInstance.array_id == array_id,
                TaskInstance.id == bindparam("task_instance_id"),
            )
            .values(distributor_id=bindparam("distributor_id"))
            .execution_options(synchronize_session=False)
        )
        session.connection().execute(update_stmt, params)

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp



# from jobmon.server.web.start import create_app  # noqa F401
from typing import Optional

from flask import Flask

from jobmon.core.configuration import JobmonConfig
from jobmon.server.web.app_factory import AppFactory  # noqa F401
from jobmon.server.web.log_config import configure_logging  # noqa F401


def get_app(config: Optional[JobmonConfig] = None) -> Flask:
    """Get a flask app based on the config. If no config is provided, defaults are used.

    Args:
        config: The jobmon config to use when creating the app.
    """
    app_factory = AppFactory(config)
    app = app_factory.get_app()
    return app



"""Task Resources Database Table."""
from sqlalchemy import Column, ForeignKey, Integer, String, Text
from sqlalchemy.orm import relationship

from jobmon.core.serializers import SerializeTaskResources
from jobmon.server.web.models import Base
from jobmon.server.web.models.task_resources_type import TaskResourcesType  # noqa F401


class TaskResources(Base):
    """The table in the database that holds all task specific resources.

    Task specific resources:
        queue_id - designated queue
        requested_resources
    """

    __tablename__ = "task_resources"

    def to_wire_as_task_resources(self) -> tuple:
        """Serialize executor task object."""
        serialized = SerializeTaskResources.to_wire(
            task_resources_id=self.id,
            queue_name=self.queue.name,
            task_resources_type_id=self.task_resources_type_id,
            requested_resources=self.requested_resources,
        )
        return serialized

    id = Column(Integer, primary_key=True)
    queue_id = Column(Integer, ForeignKey("queue.id"))
    task_resources_type_id = Column(String(1), ForeignKey("task_resources_type.id"))

    requested_resources = Column(Text, default=None)

    # ORM relationships
    queue = relationship("Queue", foreign_keys=[queue_id])
    task_resources_type = relationship(
        "TaskResourcesType", foreign_keys=[task_resources_type_id]
    )



"""Edge Database table."""
from sqlalchemy import Column, Integer, Text

from jobmon.server.web.models import Base


class Edge(Base):
    """Database Table to record edges."""

    __tablename__ = "edge"

    dag_id = Column(Integer, primary_key=True)
    node_id = Column(Integer, primary_key=True)
    # Upstream and downstream nodes are edge lists (incoming and outgoing).
    # They are python lists serialized as strings
    # with syntax: "[node_id, node_id, node_id, ...]"
    # for implementation details refer to
    # jobmon/client/swarm/workflow/clientdag.py
    # method: _insert_edges()
    upstream_node_ids = Column(Text)
    downstream_node_ids = Column(Text)



"""Workflow status database table."""
from sqlalchemy import Column, String
from sqlalchemy.orm import Session

from jobmon.core.constants import WorkflowStatus as Statuses
from jobmon.server.web.models import Base


class WorkflowStatus(Base):
    """Workflow Status database table."""

    __tablename__ = "workflow_status"

    REGISTERING = Statuses.REGISTERING
    QUEUED = Statuses.QUEUED
    ABORTED = Statuses.ABORTED
    RUNNING = Statuses.RUNNING
    HALTED = Statuses.HALTED
    FAILED = Statuses.FAILED
    DONE = Statuses.DONE
    INSTANTIATING = Statuses.INSTANTIATING
    LAUNCHED = Statuses.LAUNCHED

    id = Column(String(1), primary_key=True)
    label = Column(String(150), nullable=False)
    description = Column(String(150))


def add_workflow_statuses(session: Session) -> None:
    """Populate the workflow_status in the database."""
    statuses = [
        WorkflowStatus(
            id="A",
            label="ABORTED",
            description="Workflow encountered an error before a WorkflowRun was created.",
        ),
        WorkflowStatus(
            id="D",
            label="DONE",
            description="Workflow has completed, it finished successfully.",
        ),
        WorkflowStatus(
            id="F",
            label="FAILED",
            description="Workflow unsuccessful in one or more WorkflowRuns, "
            "no runs finished successfully as DONE.",
        ),
        WorkflowStatus(
            id="G", label="REGISTERING", description="Workflow is being validated."
        ),
        WorkflowStatus(
            id="H",
            label="HALTED",
            description="Resume was set and Workflow is shut down or the controller died "
            "and therefore Workflow was reaped.",
        ),
        WorkflowStatus(
            id="I",
            label="INSTANTIATING",
            description="Jobmon Scheduler is creating a Workflow on the distributor.",
        ),
        WorkflowStatus(
            id="O",
            label="LAUNCHED",
            description="Workflow has been created. Distributor is now controlling tasks, "
            "or waiting for scheduling loop.",
        ),
        WorkflowStatus(
            id="Q",
            label="QUEUED",
            description="Jobmon client has updated the Jobmon database, "
            "and signalled Scheduler to create Workflow.",
        ),
        WorkflowStatus(
            id="R",
            label="RUNNING",
            description="Workflow has a WorkflowRun that is running.",
        ),
    ]
    session.add_all(statuses)



"""ArgType table in the database."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship, Session

from jobmon.core import constants
from jobmon.server.web.models import Base


class ArgType(Base):
    """ArgType table in the database."""

    __tablename__ = "arg_type"

    NODE_ARG = constants.ArgType.NODE_ARG
    TASK_ARG = constants.ArgType.TASK_ARG
    OP_ARG = constants.ArgType.OP_ARG

    id = Column(Integer, primary_key=True)
    name = Column(String(255), unique=True)

    template_arg_map = relationship("TemplateArgMap", back_populates="argument_type")


def add_arg_types(session: Session) -> None:
    """Populate the arg_type table in the database."""
    types = [
        ArgType(id=1, name="NODE_ARG"),
        ArgType(id=2, name="TASK_ARG"),
        ArgType(id=3, name="OP_ARG"),
    ]
    session.add_all(types)



"""Tool version db table."""
from sqlalchemy import Column, ForeignKey, Integer
from sqlalchemy.orm import relationship

from jobmon.core.serializers import SerializeClientToolVersion
from jobmon.server.web.models import Base


class ToolVersion(Base):
    """Tool version db table."""

    __tablename__ = "tool_version"

    def to_wire_as_client_tool_version(self) -> tuple:
        """Serialize tool version object."""
        serialized = SerializeClientToolVersion.to_wire(
            id=self.id, tool_id=self.tool_id
        )
        return serialized

    id = Column(Integer, primary_key=True)
    tool_id = Column(Integer, ForeignKey("tool.id"), nullable=False)

    # ORM relationships
    tool = relationship("Tool", back_populates="tool_versions")
    task_templates = relationship("TaskTemplate", back_populates="tool_versions")



"""Database Table for Task Template Versions."""
from typing import Dict, List

from sqlalchemy import Column, ForeignKeyConstraint, Index, Integer, Text, VARCHAR
from sqlalchemy.orm import relationship

from jobmon.core.serializers import SerializeClientTaskTemplateVersion
from jobmon.server.web.models import Base
from jobmon.server.web.models.arg_type import ArgType


class TaskTemplateVersion(Base):
    """Database Table for Task Template Versions."""

    __tablename__ = "task_template_version"

    def to_wire_as_client_task_template_version(self) -> tuple:
        """Serialized Task Template Version objects."""
        # serialized = SerializeClientTool.to_wire(id=self.id, name=self.name)
        # return serialized
        id_name_map = {}
        args_by_type: Dict[str, List[str]] = {
            "node_args": [],
            "task_args": [],
            "op_args": [],
        }
        for arg_mapping in self.template_arg_map:
            id_name_map[arg_mapping.argument.name] = arg_mapping.argument.id

            if arg_mapping.arg_type_id == ArgType.NODE_ARG:
                args_by_type["node_args"].append(arg_mapping.argument.name)
            if arg_mapping.arg_type_id == ArgType.TASK_ARG:
                args_by_type["task_args"].append(arg_mapping.argument.name)
            if arg_mapping.arg_type_id == ArgType.OP_ARG:
                args_by_type["op_args"].append(arg_mapping.argument.name)

        return SerializeClientTaskTemplateVersion.to_wire(
            task_template_version_id=self.id,
            command_template=self.command_template,
            node_args=args_by_type["node_args"],
            task_args=args_by_type["task_args"],
            op_args=args_by_type["op_args"],
            id_name_map=id_name_map,
            task_template_id=self.task_template.id,
        )

    id = Column(Integer, primary_key=True)
    task_template_id = Column(Integer)
    command_template = Column(Text)
    arg_mapping_hash = Column(VARCHAR(50))

    # orm relationship
    task_template = relationship(
        "TaskTemplate", back_populates="task_template_versions"
    )
    template_arg_map = relationship(
        "TemplateArgMap", back_populates="task_template_version"
    )

    __table_args__ = (
        Index(
            "uc_ttv_composite_pk",
            "task_template_id",
            "command_template",
            "arg_mapping_hash",
            unique=True,
        ),
        Index("ix_task_template_id", "task_template_id"),
        ForeignKeyConstraint(
            ["task_template_id"], ["task_template.id"], use_alter=True
        ),
    )
